using System;
using System.IO;
using System.Net;
using System.Collections;
using System.Threading;

using Gtk;
using Glade;

using P2Pc.net;
using P2Pc.gui;
using P2Pc.util;

namespace P2Pc {

	public class P2Pc_GUI : IPL
	{
		public const string mstrMainThreadName = "Th30ne4nd0nlyMA1NTHR34D";
		public static bool mbGdkLockThread=false;

		private const string mstrVersion = "P2Pc-GUI v0.1";
		private const string mstrHelp =
		"Usage: mono P2Pc-GUI.exe [-p <port>] [-l <loglevel>]";

		// define Glade-Widgets
		[Glade.Widget] Gtk.Window main_window;
		[Glade.Widget] TextView entry1;
		[Glade.Widget] TextView textview1;
		[Glade.Widget] TreeView treeview1;
		[Glade.Widget] VPaned vpane;
		[Glade.Widget] VBox vbox2;
		[Glade.Widget] HBox hbox1;
		[Glade.Widget] HandleBox handlebox1;
		[Glade.Widget] Combo combo1;
		[Glade.Widget] Gtk.Entry combo_entry1;
		[Glade.Widget] Statusbar statusbar1;
		[Glade.Widget] CheckButton anonymize_checkbutton;
		[Glade.Widget] Image securityimage;

		private Notebook notebook1;
		private Tooltips mTooltips;

		private const string mcostrDefaultChanName = "Default";
		private string mstrPeersFile = "peers.xml";
		private bool mbAnonymize = false;
		private bool mbIsSecure = false;
		private string[] mChanCombolist = null;
		private uint mnStatusMsgContextID=0;
		private Hashtable mhtOpenChannels;
		private ClGuiChannel moActualChannel;


		public bool IsSecure {
			set{
				mbIsSecure = value;
				if(mbIsSecure){
					// put a nice picture here
				}
				else{
					// put a nice picture here
				}
			}
			get{
				return mbIsSecure;
			}
		}

		public static void Main (string[] args)
		{
			Thread.CurrentThread.Name = mstrMainThreadName;
			int port = 0;
			if (args.Length == 1){
				if (args[0].Equals("-v") || args[0].Equals("--version")){
					Console.WriteLine(mstrVersion);
				}
				else{
					Console.WriteLine(mstrHelp);
				}
				return;
			}

			for (int i=0;i<(args.Length-1);++i){
				try{
					if (args[i].Equals("-p")){
						port = Convert.ToInt32(args[++i]);
					}
					else if (args[i].Equals("-l")){
						Logger.PLevel = Convert.ToInt32(args[++i]);
					}
				}catch(Exception){
					Console.WriteLine(mstrHelp);
					return;
				}
			}
			if( 0 < port&&port < 65536 ){
				CConMan.PLocalListenPort = port;
			}
			else{
				// Start Listening on default port :)
				CConMan.PLocalListenPort = CConMan.PLocalListenPort;
			}

			Gdk.Threads.Init();
			Application.Init();

			new P2Pc_GUI();

			Application.Run();
		}

		public P2Pc_GUI ()
		{
			// create GUI from Glade-File
			/* Note that we load the XML info from the assembly instead of using
			   an external file. You don't have to distribute the .glade file if
			   you don't want */
			Glade.XML gxml = new Glade.XML ("p2pc.glade", "main_window", null);
			gxml.Autoconnect (this);
			mTooltips = new Tooltips();
			//entry1.Activated += new EventHandler(text_input);

			// add notebook to GUI (for tabs of opened channels)
			notebook1 = new Notebook();
			notebook1.SwitchPage += new SwitchPageHandler(page_change);
			// notebook1.Scrollable = true; // <-- propably buggy...
			notebook1.SetSizeRequest(250,150);
			vbox2.PackStart(notebook1,true,true,0);
			vbox2.ReorderChild(notebook1, 2);
			notebook1.Show();
			// vpane.Pack1(notebook1,true,false);

			// register delegate for p2pc-netinfos
			CConMan.EvNetInfoChanged +=
				new DNetInfoChangedEventHandler(netinfo_changed);

			// create list of available channels and open Default channel
			update_channelcombo(null);
			mhtOpenChannels = new Hashtable();
			moActualChannel = new ClGuiChannel((IChannel)CConMan.PKnownChannels[""], main_window);
			mhtOpenChannels.Add(mcostrDefaultChanName,moActualChannel);
			notebook1.AppendPage(moActualChannel.getChanWidget(),
								 new Label(mcostrDefaultChanName));
			combo1.Entry.Text = mcostrDefaultChanName;

			// show GUI
			main_window.ShowAll();
		}

		public void setNodesFile(string strNodesFile){
			if(strNodesFile!=null){
				mstrPeersFile = strNodesFile;
			}
		}

		public void netinfo_changed(object sender, NetInfoChangedEventArgs e){
			// Console.WriteLine("net evt call by: "+Thread.CurrentThread.Name);
			if(mbGdkLockThread){
				throw new SynchronizationLockException("Gdk: Already entered!");
			}
			if( (Thread.CurrentThread.Name!=mstrMainThreadName)){
				mbGdkLockThread=true;
				Gdk.Threads.Enter();
			}

			switch(e.changedNetInfo){
			case NetInfoEventType.channel:
				Console.WriteLine("Channel change Event!"+(IChannel)sender);
				if(sender!=null){
					update_channelcombo((IChannel)sender);
				}
				break;
			case NetInfoEventType.user:
				//Console.WriteLine("User change Event!"+(IUser)sender);
				break;
			case NetInfoEventType.node:
				//Console.WriteLine("Node change Event! Node: "+(INode)sender);
				break;
			case NetInfoEventType.generic:
				//Console.WriteLine("Generic change Event!");
				break;
			default:
				Console.WriteLine("Unknown net change Event!");
				break;
			}

			if(mbGdkLockThread){
				Gdk.Threads.Leave();
				mbGdkLockThread = false;
			}
		}


		public void quit(object o, DeleteEventArgs args){
			args.RetVal = true;
			quit();
		}

		public void quit(System.Object obj, EventArgs e){
			quit();
		}

		public void quit(){
			foreach(INode oNodeIter in CConMan.PKnownNodes.Values){
				disconnect_node(oNodeIter);
			}
			try{
				CConMan.SaveKnownNodes(new FileStream(mstrPeersFile,FileMode.Create));
			}
			catch(Exception exc){
				ClMsgDias.Err("Error saving Peers to file\n\""+mstrPeersFile+"\"!",
							  main_window);
			}
			Application.Quit ();
			System.Console.WriteLine("Quit.");
		}

		public void select_channel(object source, EventArgs args){
			select_channel(combo1.Entry.Text);
		}

		public void select_channel(string strChannel){
			if(strChannel=="") return;
			if(strChannel==mcostrDefaultChanName
			   ||
			   mhtOpenChannels.Contains(strChannel))
			{
				notebook1.CurrentPage = getTabNumber(strChannel);
				entry1.HasFocus = true;
				main_window.ShowAll();
			}
		}

		public void join_channel(object source, EventArgs args){
			join_channel(combo1.Entry.Text);
		}

		public void join_channel(string strChannel){
			if(strChannel=="") strChannel = mcostrDefaultChanName;
			if(strChannel==mcostrDefaultChanName
			   ||
			   mhtOpenChannels.Contains(strChannel))
			{
				notebook1.CurrentPage = getTabNumber(strChannel);
				ClMsgDias.Info("You are alreadey Member in Channel "+
							   strChannel+"!", main_window);
			}
			else{
				try{
					IChannel oChan = (IChannel)CConMan.addToKnownChannels(strChannel);
					oChan.join();
					ClGuiChannel oChannel = new ClGuiChannel(oChan, main_window);
					mhtOpenChannels.Add(strChannel,oChannel);
					notebook1.AppendPage(oChannel.getChanWidget(),
										 new Label(strChannel));
					notebook1.CurrentPage = getTabNumber(strChannel);
				}
				catch(Exception exc){
					ClMsgDias.Warn("Could not join channel:\n"+exc,
								   main_window);
					return;
				}
			}
			entry1.HasFocus = true;
			main_window.ShowAll();
		}

		public void close_channel(object source, EventArgs args){
			if(moActualChannel.Channel.PName==""){
				ClMsgDias.Info("This is the Default Channel.\n"+
							   "You don't have to join or leave it.",
							   main_window);
				return;
			}
			string strChannel =
				notebook1.GetTabLabelText(notebook1.GetNthPage(notebook1.CurrentPage));
			notebook1.RemovePage(notebook1.CurrentPage);
			mhtOpenChannels.Remove(strChannel);
			//System.Console.WriteLine("closed: {0}",i);
		}

		public void text_input(object source, EventArgs args){
			// System.Console.WriteLine("Text entered: " + (String)entry1.Text);
			if(CConMan.FbOnlyNode){
				ClMsgDias.Warn("Can't post Message.\n" +
							  "Set yourself a Nick first\n" +
							  "(using Preferences Menu)",
							  main_window);
			}
			else{
				if(moActualChannel.Channel.PName=="" && mbAnonymize){
					try{
						CConMan.sendAnonMsg(entry1.Buffer.Text);
					}catch(Exception exc){
						ClMsgDias.Warn("Could not send anonym message:\n"+exc,
									   main_window);
					}
				}
				else{
					try{
						moActualChannel.sendMsg(CConMan.PLocalNickName, entry1.Buffer.Text);
					}catch(Exception exc){
						ClMsgDias.Warn("Could not send message:\n"+exc,
									   main_window);
					}
				}
				entry1.Buffer.Text = "";
			}
		}

		public void anonymize_toggle(System.Object obj, EventArgs e){
			mbAnonymize = !mbAnonymize;
			IsSecure = mbAnonymize;
			entry1.HasFocus = true;
		}

		public void show_preferences(System.Object obj, EventArgs e){
			ClPreferencesDialog oPDia =
				new ClPreferencesDialog(main_window,
										CConMan.PLocalNickName,
										CConMan.PLocalListenPort,
										mstrPeersFile,
										this);
		}

		public void create_channel_dia(System.Object obj, EventArgs e){
			if(CConMan.FbOnlyNode){
				ClMsgDias.Warn("Can't create channel.\n" +
							  "Set yourself a Nick first\n" +
							  "(using Preferences Menu)",
							  main_window);
			}
			else{
				ClCreateChanDialog oCChanDia = new ClCreateChanDialog(main_window,
								this);
			}
		}

		public void create_channel(string strChanName, string strTopic, bool bIsClosed){
			if(strChanName!=null){
				if(CConMan.PKnownChannels.Contains(strChanName)){
					ClMsgDias.Warn("Channel with that Name exists!",main_window);
				}
				else{
					IChannel oChan;
					try{
						oChan = (IChannel)CConMan.createChannel(strChanName,bIsClosed);
						oChan.join();
						ClGuiChannel oChannel = new ClGuiChannel(oChan, main_window);
						mhtOpenChannels.Add(strChanName,oChannel);
						notebook1.AppendPage(oChannel.getChanWidget(),
											 new Label(strChanName));
						notebook1.CurrentPage = getTabNumber(strChanName);
					}catch(Exception exc){
						ClMsgDias.Warn("Could not create channel:\n"+exc,
									   main_window);
						return;
					}
					update_channelcombo(oChan);
					select_channel(strChanName);
					//join_channel(strChanName); is done by CConMan.createChannel now !
					entry1.HasFocus = true;
					main_window.ShowAll();
				}
			}
		}

		public void changetopic_dia(System.Object obj, EventArgs e){
			ClChangeTopicDialog oCTDia =
				new ClChangeTopicDialog(main_window,
										this,
										moActualChannel.Channel,
										moActualChannel.Channel.PTopic);
		}

		public void change_topic(IChannel oChannel, string strNewTopic){
			try{
				string strTopicAndInfo = strNewTopic;
				if(oChannel.PTopic!=null) strTopicAndInfo = oChannel.PTopic;
				strTopicAndInfo =
					( (strTopicAndInfo=="")? "(Topic not defined)" : "Topic: \""+strTopicAndInfo ) +
					"\"\nChannel is " + ( (oChannel.PbClosed)? "closed." : "open." );
				mTooltips.SetTip(combo_entry1, strTopicAndInfo, null);
				oChannel.PTopic = strNewTopic;
			}catch(Exception exc){
				ClMsgDias.Warn("Could not send accept message to user:\n"+exc,
							   main_window);
			}
		}

		public void acceptuser_dia(System.Object obj, EventArgs e){
			if(!moActualChannel.Channel.PbClosed){
				ClMsgDias.Info("All users may enter this channel:\n"+
							   "it's not closed!",
							   main_window);
			}
			else{
				ClAcceptDialog oAccDia =
					new ClAcceptDialog(main_window,
									   this,
									   moActualChannel.Channel.PName);
			}
		}

		public void accept_user(string strUser2Acc, string strClosedChannel){
			try{
				((IChannel)CConMan.PKnownChannels[strClosedChannel]).acceptUser(
						(IUser)CConMan.PKnownUsers[strUser2Acc] );
			}catch(Exception ex2){
				ClMsgDias.Warn("Could not send accept message to user:\n"+ex2,
							   main_window);
			}
		}


		public void show_logwindow(System.Object obj, EventArgs e){
			ClLogDialog oLogDialog = new ClLogDialog(main_window);
		}

		public bool add_node(string strHost, int port){
			CNode cn;
			IPAddress ip=null;
			bool bReturn=true;
			try{
				ip = IPAddress.Parse(strHost);
			}catch(Exception exc){		//no IP, have look it up
				try{
					ip = Dns.GetHostByName(strHost).AddressList[0];
				}catch(Exception ex2){
					ClMsgDias.Warn("Failed to look up hostname:\n"+ex2,
								   main_window);
					bReturn = false;
				}
			}
			if(ip!=null){
				cn = CConMan.addToKnownNodes(new IPEndPoint(ip,port));
				cn.openLink();
			}
			return bReturn;
		}

		public void connect_node(INode oNode){
			try{
				oNode.openLink();
			}
			catch(Exception exc){
				ClMsgDias.Warn("Could not open link:\n"+exc,
							   main_window);
			}
		}

		public void disconnect_node(INode oNode){
			try{
				oNode.closeLink();
			}
			catch(Exception exc){
				ClMsgDias.Warn("Could not close link:\n"+exc,
							   main_window);
			}
		}

		public bool remove_node(string strHost, int port){
			CNode cn;
			IPAddress ip=null;
			bool bReturn=true;
			try{
				ip = IPAddress.Parse(strHost);
			}catch(Exception exc){		//no IP, have look it up
				try{
					ip = Dns.GetHostByName(strHost).AddressList[0];
				}catch(Exception ex2){
					Console.WriteLine("failed to look up hostname:"+ex2,
									  main_window);
					bReturn = false;
				}
			}
			if(ip!=null){
				cn = CConMan.addToKnownNodes(new IPEndPoint(ip,port));
				remove_node(cn);
			}
			return bReturn;
		}

		public void remove_node(INode oNode){
			try{
				oNode.closeLink();
				CConMan.remNode(oNode);
			}catch(Exception exc){
				ClMsgDias.Warn("Failed to remove Peer:\n"+exc,
							   main_window);
			}

		}

		public void show_about(System.Object obj, EventArgs e){
			ClAboutDialog about = new ClAboutDialog(main_window);
		}

		public void connect_peers(System.Object obj, EventArgs e){
			connect();
		}

		public void connect(){
			write_status("Connecting peers ...");
			try{
				if(File.Exists(mstrPeersFile)){
					CConMan.LoadKnownNodes(new FileStream(mstrPeersFile,FileMode.Open));
				}
				else throw new Exception();
			}
			catch(Exception exc){
				ClMsgDias.Warn("Error opening/reading Node File "+mstrPeersFile+"!",
							  main_window);
			}
		}

		public void disconnect_peers(System.Object obj, EventArgs e){
			write_status("Disconnecting peers ...");
		}

		private void update_channelcombo(IChannel oChan){
			if(oChan==null
			   ||
			   mChanCombolist==null
			   ||
			   Array.BinarySearch(mChanCombolist, oChan.PName)<0)
			{
				mChanCombolist = new string[CConMan.PKnownChannels.Count];
				int i=0;
				foreach(IChannel oChanIter in CConMan.PKnownChannels.Values){
					if(oChanIter.PName!="") mChanCombolist[i++] = oChanIter.PName;
					else mChanCombolist[i++] = mcostrDefaultChanName;
				}
				Array.Sort(mChanCombolist);
				combo1.PopdownStrings = mChanCombolist;
				if(oChan!=null) combo_entry1.Text = oChan.PName;
				main_window.ShowAll();
			}
		}

		private int getTabNumber(string str){
			int retVal=0;
			for(int i=0; i<notebook1.NPages; ++i){
				if(System.String.Compare(str,
										 notebook1.GetTabLabelText(notebook1.GetNthPage(i)))
				   == 0)
				{
					retVal=i;
					break;
				}
			}
			return retVal;
		}

		private void page_change(object source, SwitchPageArgs args){
			string strChannel =
				notebook1.GetTabLabelText(notebook1.GetNthPage(notebook1.CurrentPage));
			moActualChannel = (ClGuiChannel)mhtOpenChannels[strChannel];
			if(moActualChannel.Channel.PName=="") IsSecure = mbAnonymize;
			if(moActualChannel.Channel.PbClosed) IsSecure = true;
			else IsSecure = false;
			string strTopicAndInfo = "";
			if(moActualChannel.Channel.PTopic!=null)
				strTopicAndInfo = moActualChannel.Channel.PTopic;
			strTopicAndInfo =
				( (strTopicAndInfo=="")? "(Topic not defined)" : "Topic: \""+strTopicAndInfo ) +
				"\"\nChannel is " + ( (moActualChannel.Channel.PbClosed)? "closed." : "open." );
			if(strChannel==mcostrDefaultChanName){
				anonymize_checkbutton.Active = mbAnonymize;
				anonymize_checkbutton.Sensitive = true;
			}
			else{
				anonymize_checkbutton.Sensitive = false;
			}
			combo_entry1.Text = strChannel;
			mTooltips.SetTip(combo_entry1,
							 strTopicAndInfo,
							 null);
			entry1.HasFocus = true;
		}

		private void write_status(string str){
			statusbar1.Pop(mnStatusMsgContextID);
			mnStatusMsgContextID = statusbar1.Push(statusbar1.GetContextId(str),str);
		}
	}
}
