/* class P2Pc_Node - the main class for the Console only Frontend of P2Pc
 * uses the library (CConMan) and has many cool features ;)
 * just starts the listening socket waits for User input, parses it and calls the CConMan
 * 
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 24.06.2004
 * Time: 22:51
 * 
 */

using System;
using System.Threading;
using System.Net;
using System.IO;
using Mono.Security;
using System.Security.Cryptography;
using Mono.Security.Cryptography;

using P2Pc.net;
using P2Pc.util;

namespace P2Pc
{
	/// <summary>
	/// P2Pc:Node-only daemon	
	/// </summary>
	public class P2Pc_Node
	{
		private static string FstrVersion =
    		"P2Pc-Node v0.2:\nNode only (console)-Frontend for the IBR-P2P Chat Protocol Version2";
    		
		private static string FstrHelp =
    		"Usage: mono P2Pc-Node.exe [options]\n\n"
    		+"where options include:\n-h,--help\tdisplays this help\n"
    		+"-p <port>\tsets the port to listen on"
		+"-l <level>\tsets the log level";
		private static string FstrCommands =
			"Understood commands are:\n"
			+ "open <host:port>\topens link to node\n"
			+ "close <host:port>\tcloses link to node\n"
			+ "s,stat\t\t\tprints out all NetInfo\n"
			+ "l [l]\t\t\tsets log-level to l (0..5) (default:1)\n"
			+ "log2 [file]\t\tsets output for Logger to file (default:P2Pc-Node.log)\n"
			+ "p [port]\t\tsets the local listen port (default:4321)\n"
			+ "anon [msg]\t\tsends an anonymous message\n"
			+ "q,exit,bye,quit\t\texits\n"
			+ "save [file]\t\tsaves local Known Nodes to <file> (default:KnownNodes.xml)\n"
			+ "load [file]\t\tloads local Known Nodes from <file> (default:KnownNodes.xml)\n"
			+ "nick <name>\t\tsets the local nick name and tries to load cert/key\n"
			+ "nick\t\t\tresets the local NickName and cert\n"
			+ "---below commands only work if nick is set---\n"
			+ "cert <user>\t\tsends the local cert to user\n"
			+ "create <channel>\tcreates a channel\n"
			+ "cclosed <channel>\tcreates a closed channel\n"
			+ "join <channel>\t\tjoins a given channel\n"
			+ "leave <channel>\t\tleaves a given channel\n"
			+ "acc <channel> <user>\taccepts user to closed channel <channel>\n"
			+ "top <channel> <topic>\tsets the topic for <channel>\n"
			+ "msgto <channel> <text>\tsends msg to channel\n"
			+ "msg <text>\t\tsends msg to omnipresent, i.e. \"\"-channel";
		
		public P2Pc_Node(int pPort)
		{
			Logger.Log("Daemon",0,"starting up...");
			Logger.Log("Daemon",0,"Listening on port "+pPort);
			CConMan.PLocalListenPort = pPort;
			
			//set Prio low
			Thread.CurrentThread.Priority = ThreadPriority.Lowest;
			//and do nothing... just yield our timeslice
			string C;
			bool nickSet = false;
			//reg. local "" channel events
			((CChannel)CConMan.PKnownChannels[""]).EvChannelEvent += new DChannelEventHandler(myChannelEvent);
				
			while(true){
				try { Thread.Sleep(100); }
				catch(Exception){ }
				//Look if theres something to do
				C = Console.ReadLine();
				//what should we do?
				
				if (C.Equals("") || C.Equals("help") || C.Equals("h")){
					Console.WriteLine(FstrCommands);
					Console.WriteLine("Nick is"+((nickSet)?"":" not")+" set"+((nickSet)?" to "+CConMan.PLocalNickName:"")+"\n");
				}else if (C.StartsWith("open ")){
				try{
					CNode cn;
					string fullip = C.Substring(5);
					int idx = fullip.LastIndexOf(':');
					int port = Int32.Parse(fullip.Substring(idx+1));
					string ip =  fullip.Substring(0,idx);
					IPAddress ipa;
					try{
						ipa = IPAddress.Parse(ip);
					}catch(Exception){		//no IP, have look it up
						try{
							ipa = Dns.GetHostByName(ip).AddressList[0];
						}catch(Exception e){
							Console.WriteLine("failed to look up hostname:"+e);
							ipa=null;
						}
 					}
 					if (ipa!=null){
 						cn = CConMan.addNode(new IPEndPoint(ipa,port));
 						cn.openLink();
 					}
				}catch(Exception e){
					Console.WriteLine("Could not open link:"+e);
				}
				}else if (C.StartsWith("close ")){
				try{
					CNode cn;
					string fullip = C.Substring(6);
					int idx = fullip.LastIndexOf(':');
					int port = Int32.Parse(fullip.Substring(idx+1));
					string ip =  fullip.Substring(0,idx);
					IPAddress ipa;
					try{
						ipa = IPAddress.Parse(ip);
					}catch(Exception){		//no IP, have look it up
						try{
							ipa = Dns.GetHostByName(ip).AddressList[0];
						}catch(Exception e){
							Console.WriteLine("failed to look up hostname:"+e);
							ipa=null;
						}
 					}
 					if (ipa!=null){
 						cn = CConMan.addToKnownNodes(new IPEndPoint(ipa,port));
 						cn.closeLink();
 					}
				}catch(Exception e){
					Console.WriteLine("Could not close link:"+e);
				}
					
				}else if (C.StartsWith("nick ")){
				try{	
					if (C.Length < 7){
						CConMan.PLocalNickName = null;
						CConMan.PLocalX509Cert = null;
						CConMan.PLocalRSA = null;
						nickSet = false;
					}else{
						string sNick = C.Substring(5);
						if (!File.Exists(sNick+"-cert.der") || !File.Exists(sNick+"-key.pem.der"))
							Console.WriteLine("key or cert file not present in current dir");
						else{
							CConMan.PLocalNickName = sNick;
							CConMan.PLocalX509Cert = Load(sNick+"-cert.der");
							try{
	 							ASN1 privateKey = new ASN1(Load(sNick+"-key.pem.der"));
								RSA rsa = PKCS8.PrivateKeyInfo.DecodeRSA(privateKey[2].Value);
								if (rsa!=null){
									CConMan.PLocalRSA = rsa;
									nickSet = true;
								}
							}catch(Exception e){
								Logger.Log("Daemon",1,"failed setting local RSA-Key",e);
							}
						}
					}
				}catch(Exception e){
					Console.WriteLine("Could not set nick:"+e);
				}
				
				}else if (C.StartsWith("create ") && nickSet){
					try{
						IChannel cc = CConMan.createChannel(C.Substring(7));
						//cc.join();
						cc.EvChannelEvent += new DChannelEventHandler(myChannelEvent);
					}catch(Exception e){
						Console.WriteLine("Could not create channel:"+e);
					}
				
				}else if (C.StartsWith("cert ") && nickSet){
					try{
						CConMan.sendCertificateMsgTo(C.Substring(5));
					}catch(Exception e){
						Console.WriteLine("Could send cert:"+e);
					}	
				}else if (C.StartsWith("cclosed ") && nickSet){
					try{
						IChannel cc = CConMan.createChannel(C.Substring(8),true);
						cc.EvChannelEvent += new DChannelEventHandler(myChannelEvent);
					}catch(Exception e){
						Console.WriteLine("Could not create closed channel:"+e);
					}
				}else if (C.StartsWith("join ") && nickSet){
					try{
						IChannel cc = CConMan.addToKnownChannels(C.Substring(5));
						cc.join();
						cc.EvChannelEvent += new DChannelEventHandler(myChannelEvent);
					}catch(Exception e){
						Console.WriteLine("Could not join channel:"+e);
					}
				}else if (C.StartsWith("leave ") && nickSet){
					try{
						IChannel cc = CConMan.addToKnownChannels(C.Substring(6));
						cc.leave();
						cc.EvChannelEvent -= new DChannelEventHandler(myChannelEvent);
					}catch(Exception e){
						Console.WriteLine("Could not leave channel:"+e);
					}
				}else if (C.StartsWith("msgto ") && nickSet){
					try{
						C = C.Substring(6);
						int idx = C.IndexOf(' ');
						((IChannel)CConMan.PKnownChannels[C.Substring(0,idx)]).sendMessage(C.Substring(idx+1));
					}catch(Exception e){
						Console.WriteLine("Could not send msg to channel:"+e);
					}
				}else if (C.StartsWith("msg ") && nickSet){
					try{
						C = C.Substring(4);
						((IChannel)CConMan.PKnownChannels[""]).sendMessage(C);
					}catch(Exception e){
						Console.WriteLine("Could not send msg:"+e);
					}
				}else if (C.StartsWith("anon")){
					try{
						string msg="";
						if (C.Length > 5) msg=C.Substring(5);
						CConMan.sendAnonMsg(msg);
					}catch(Exception e){
						Console.WriteLine("Could not send anon msg:"+e);
					}
				}else if (C.StartsWith("acc ") && nickSet){
					try{
						C = C.Substring(4);
						int idx = C.IndexOf(' ');
						((IChannel)CConMan.PKnownChannels[C.Substring(0,idx)]).acceptUser((IUser)CConMan.PKnownUsers[C.Substring(idx+1)]);
					}catch(Exception e){
						Console.WriteLine("Could not accept user to channel:"+e);
					}
				}else if (C.StartsWith("top ") && nickSet){
					try{
						C = C.Substring(4);
						int idx = C.IndexOf(' ');
						IChannel iChan = ((IChannel)CConMan.PKnownChannels[C.Substring(0,idx)]);
						if (iChan.PbLocalUserMember)
							iChan.PTopic = C.Substring(idx+1);
					}catch(Exception e){
						Console.WriteLine("Could not set topic for channel:"+e);
					}
				}else if (C.Equals("s") || C.Equals("stat")){
					try{
						Console.WriteLine("Nodes:");
						foreach (CNode cn in CConMan.PKnownNodes.Values) Console.WriteLine("  "+cn);
						Console.WriteLine("Channels:");
						foreach (CChannel cc in CConMan.PKnownChannels.Values) Console.WriteLine(" "+((cc.PbLocalUserMember)?"@":"")+cc);
						Console.WriteLine("Users:");
						foreach (CUser cu in CConMan.PKnownUsers.Values) Console.WriteLine("  "+cu);
						Console.WriteLine(CKnownMsgs.PCount+" MIDs in Table");
					}catch(Exception e){
						Console.WriteLine("Could not display stats:"+e);
					}
				}else if (C.StartsWith("log2")){
					try{
						string filen = "P2Pc-Node.log";
						if (C.Length > 5) filen=C.Substring(5);
						if (!File.Exists(filen)) Logger.PStream = new StreamWriter(File.Create(filen));
						else Logger.PStream = new StreamWriter(filen);
					}catch(Exception e){
						Console.WriteLine("Could not set log file:"+e);
					}
				}else if (C.StartsWith("load")){
					try{
						string filen = "KnownNodes.xml";
						if (C.Length > 5) filen=C.Substring(5);
						if (File.Exists(filen))
							CConMan.LoadKnownNodes(new FileStream(filen,FileMode.Open));
					}catch(Exception e){
						Console.WriteLine("Could not load known nodes:"+e);
					}
				}else if (C.StartsWith("save")){
					try{
						string filen = "KnownNodes.xml";
						if (C.Length > 5) filen=C.Substring(5);
						CConMan.SaveKnownNodes(new FileStream(filen,FileMode.Create));
					}catch(Exception e){
						Console.WriteLine("Could not save known nodes:"+e);
					}
				}else if (C.StartsWith("l ")){
					try{
						Logger.PLevel = Int32.Parse(C.Substring(2));
					}catch(Exception e){
						Console.WriteLine("Could not set log level:"+e);
					}
				}else if (C.Equals("l")){
					Logger.PLevel = 1;
				}else if (C.Equals("p")){
					CConMan.PLocalListenPort = 4321;
				}else if (C.StartsWith("p ")){
					try{
						CConMan.PLocalListenPort = Int32.Parse(C.Substring(2));
					}catch(Exception e){
						Console.WriteLine("Could not set port:"+e);
					}
				}else if (C.Equals("q") || C.Equals("quit") || C.Equals("bye") || C.Equals("exit") ){
					try{
						CConMan.Destroy();
					}catch(Exception e){
						Console.WriteLine("Error destroying ConMan:"+e);
					}
					break;
				}else{
					Console.WriteLine("Cannot understand/execute Command: \""+C+"\"");
					Console.WriteLine(FstrCommands);
					Console.WriteLine("Nick is"+((nickSet)?"":" not")+" set"+((nickSet)?" to "+CConMan.PLocalNickName:"")+"\n");
				}
			}
			Console.WriteLine("Exiting(0)");
					
		}
		private void myChannelEvent(object sender, ChannelEventArgs e){
			string s="anon";
			if (e.affectedUser!=null) s = e.affectedUser.PNick;
			Console.Write("(" + ((CChannel)sender).PName + ":" + s + ") ");
			if (e.type == ChannelEventType.Message) Console.WriteLine(e.text);
			else if (e.type == ChannelEventType.TopicSet) Console.WriteLine("Set topic to: "+e.text);
			else Console.WriteLine(e.type);
		}
		private byte[] Load (string filename)
		{
			byte[] data = null;
			using (FileStream fs = File.OpenRead (filename)) {
				data = new byte [fs.Length];
				fs.Read (data, 0, data.Length);
			}
			return data;
		}
		
		public static void Main (string[] args)
		{
			int port = 4321;
			
			for (int i=0;i<args.Length;i++){
				if (args[i].Equals("-p"))
					port = Int32.Parse(args[++i]);
				else if (args[i].Equals("-l"))
					Logger.PLevel = Int32.Parse(args[++i]);
			}
			if (args.Length == 1){
				if (args[0].Equals("-h") || args[0].Equals("--help"))
					Console.WriteLine(FstrHelp);
				else if (args[0].Equals("-V"))
			    	Console.WriteLine(FstrVersion);
				
			}else new P2Pc_Node (port);
			
		}
	}
}
