using System;
using System.Threading;

using Gtk;
using Glade;

using P2Pc.net;

namespace P2Pc.gui {

	public class ClGuiChannel
	{
		IChannel moChannel;
		HBox mHBox;
		ScrolledWindow mSWTextView;
		ScrolledWindow mSWTreeView;
		TextView mTextView;
		TreeView mTreeView;
		TreeStore mTreeStore;
		Window moParentWin;

		public ClGuiChannel(IChannel oChannel, Window parent)
		{
			moChannel = oChannel;
			moParentWin = parent;
			mTextView = new TextView();
			mTextView.Editable = false;
			mTextView.CursorVisible = false;
			mTreeStore = new TreeStore(typeof(string));
			mTreeView = new TreeView(mTreeStore);
			if(oChannel!=null){
				if(oChannel.PUsers!=null){
					foreach(IUser oUser in oChannel.PUsers){
						if(oUser.PNick!=null) mTreeStore.AppendValues(oUser.PNick);
					}
				}
			}

			mTreeView.AppendColumn ("Members", new CellRendererText(),"text",0);
			mSWTreeView = new ScrolledWindow();
			mSWTreeView.AddWithViewport(mTreeView);
			mSWTextView = new ScrolledWindow();
			mSWTextView.SetSizeRequest(250,50);
			mSWTextView.AddWithViewport(mTextView);
			mHBox = new HBox();
			// TODO: mHPane.ComputePosition(int,int,int);
			mHBox.PackStart(mSWTextView,true,true,0);
			mHBox.PackStart(mSWTreeView,true,true,0);
			moChannel.EvChannelEvent += 
				new DChannelEventHandler(channelEventMethod);
		}

		public IChannel Channel{
			get{
				return moChannel;
			}
		}

		public Widget getChanWidget(){
			return mHBox;
		}
		
		private void channelEventMethod(object sender, ChannelEventArgs e){
			// Console.WriteLine("chan evt call by: "+Thread.CurrentThread.Name);
			if(P2Pc_GUI.mbGdkLockThread){
				throw new SynchronizationLockException("Gdk: Already entered!");
			}
			if(Thread.CurrentThread.Name!=P2Pc_GUI.mstrMainThreadName)
			{
				P2Pc_GUI.mbGdkLockThread=true;
				Gdk.Threads.Enter();
			}

			string strSender="*Anonym*@nowhere";
			string strMsg;
			if (e.affectedUser!=null) strSender = e.affectedUser.PNick;
			switch(e.type){
			case ChannelEventType.Message:
				strMsg = e.text;
				break;
			case ChannelEventType.UserWantsToJoin:
				//TODO
				strMsg = "[User wants to join this channel.]";
				break;
			case ChannelEventType.UserJoined:
				strMsg = "[User joined this channel.]";
				userupdate();
				break;
			case ChannelEventType.UserLeft:
				strMsg = "[User has left this channel.]";
				userupdate();
				break;
			default:
				strMsg = e.type.ToString();
				break;
			}
			writeMsg(strSender, strMsg);

			if(P2Pc_GUI.mbGdkLockThread){
				Gdk.Threads.Leave();
				P2Pc_GUI.mbGdkLockThread = false;
			}
		}

		private void writeMsg(string strUser, string strMsg){
			//TODO: sometimes the msg does not show up, only after another click in the textView...
			mTextView.Buffer.Text += 
				("[" +strUser.Substring(0,strUser.IndexOf("@")) + "] " +
				 String.Format("{0:dddd, dd.MM.yy hh:mm:ss}",DateTime.Now) + "\n" +
				 strMsg + "\n");
			//mTextView.FocusVadjustment = mSWTextView.Vadjustment;
			mTextView.ScrollToIter(mTextView.Buffer.EndIter,0.0,true,0,0);
			mSWTextView.Vadjustment.Value = mSWTextView.Vadjustment.Upper;
			mSWTextView.ShowAll();
		}

		public void sendMsg(string strSender, string strMsg){
			moChannel.sendMessage(strMsg);
		}
		
		public void userupdate(){
			mTreeStore.Clear();
			foreach(IUser oUser in moChannel.PUsers){
				mTreeStore.AppendValues(oUser.PNick);
			}
		}
	}
}
