using System;
using System.Net;

using Gtk;
using Glade;

using P2Pc.net;

namespace P2Pc.gui {

	public class ClNodeInfoDialog
	{
		[Glade.Widget] Gtk.Dialog nodeinfo_dialog;
		[Glade.Widget] Label hostlabel;
		[Glade.Widget] Label iplabel;
		[Glade.Widget] Label portlabel;
		[Glade.Widget] Label linktimelabel;
		[Glade.Widget] Label lastmsglabel;
		[Glade.Widget] Label receivedlabel;
		[Glade.Widget] Label sendlabel;
		[Glade.Widget] Button connectbutton;
		IPL moPL;
		DPeerListUpdateMethod mUpdatePeersMethod;
		INode moNode;

		public ClNodeInfoDialog(Gtk.Window oParentWindow,
								INode oNode,
								IPL oPL,
								DPeerListUpdateMethod oUpdatePeersMethod)
		{
			moPL = oPL;
			mUpdatePeersMethod = oUpdatePeersMethod;
			moNode = oNode;
			Glade.XML gxmlNodeInfoDia = new Glade.XML(null, "p2pc.glade", "nodeinfo_dialog", null);
			gxmlNodeInfoDia.Autoconnect(this);
			nodeinfo_dialog.TransientFor = oParentWindow;

			IPHostEntry host=null;
			string strHost=null;
			try{
				host = Dns.GetHostByAddress(oNode.PNodeName.Address);
			}catch(Exception e){
				strHost = "(hostname unknown)";				
			}
			if(host!=null) strHost = host.HostName;
			hostlabel.Text = strHost;
			iplabel.Text = oNode.PNodeName.Address.ToString();
			portlabel.Text = oNode.PNodeName.Port.ToString();

			if(oNode.PLink!=null && oNode.PLink.PbConnected){
				connectbutton.Label = "Disconnect";
				linktimelabel.Text = String.Format("{0:u}",oNode.PLink.PConnectedSince);
				lastmsglabel.Text = String.Format("{0:u}",oNode.PLastRecvCM);
				receivedlabel.Text = oNode.PLink.PTrafficRx.ToString() + " bytes";
				sendlabel.Text = oNode.PLink.PTrafficSx.ToString() + " bytes";
			}
			else{
				connectbutton.Label = "Connect";
				linktimelabel.Text = "(not connected)";
				lastmsglabel.Text = @"%";
				receivedlabel.Text = @"%";
				sendlabel.Text = @"%";
			}
			nodeinfo_dialog.ShowAll();
		}

		protected void leave_info(object sender, EventArgs e)
		{
			nodeinfo_dialog.Hide();
		}
		protected void change_connection(object sender, EventArgs e)
		{
			if(moNode.PLink!=null && moNode.PLink.PbConnected){
				moPL.disconnect_node(moNode);
			}
			else {
				moPL.connect_node(moNode);
			}
			nodeinfo_dialog.Hide();
		}
		protected void remove_node(object sender, EventArgs e)
		{
			mUpdatePeersMethod();
			moPL.remove_node(moNode);
			nodeinfo_dialog.Hide();
		}
	}
}
