using System;
using System.Net;
using System.IO;
using System.Security.Cryptography;
using Mono.Security;
using Mono.Security.Cryptography;

using Gtk;
using Glade;

using P2Pc.util;
using P2Pc.net;

namespace P2Pc.gui {

	public delegate void DPeerListUpdateMethod();

	public class ClPreferencesDialog
	{
		[Glade.Widget] Gtk.Dialog preferences_dialog;
		[Glade.Widget] Entry nick_entry;
		[Glade.Widget] Entry port_entry;
		[Glade.Widget] Entry peerfile_entry;
		[Glade.Widget] TreeView peers_treeview;
		TreeStore mTreeStore;
		IPL moPL;

		public ClPreferencesDialog(Gtk.Window oParentWindow,
								   string strNick,
								   int nPort,
								   string strPeerFile,
								   IPL oPL)
		{
			moPL = oPL;
			Glade.XML gxmlpreferences = new Glade.XML(null, "p2pc.glade", "preferences_dialog", null);
			if(gxmlpreferences!=null){
				gxmlpreferences.Autoconnect(this);
				// preferences_dialog.DefaultResponse = ResponseType.Close;
				preferences_dialog.TransientFor = oParentWindow;
				if(nick_entry!=null && strNick!=null)
					nick_entry.Text = strNick;
				if(port_entry!=null)
					port_entry.Text = nPort.ToString();
				if(peerfile_entry!=null && strPeerFile!=null)
					peerfile_entry.Text = strPeerFile;
			}
			mTreeStore = new TreeStore(typeof(string));
			peers_treeview.Model = mTreeStore;
			peers_treeview.AppendColumn ("Peers", new CellRendererText(),"text",0);
			update_peers();
		}

		public void set_peerfile(object sender, EventArgs e){
			moPL.setNodesFile(peerfile_entry.Text);
		}

		public void read_peers(object sender, EventArgs e){
			moPL.connect();
		}

		public void update_peers(object sender, EventArgs e){
			update_peers();
		}

		private void update_peers(){
			mTreeStore.Clear();
			foreach(CNode oPeerIter in CConMan.PKnownNodes.Values){
				if (!oPeerIter.FbLocal){
					IPEndPoint oIP = (IPEndPoint)oPeerIter.PNodeName;
					mTreeStore.AppendValues(oIP.ToString());
				}
			}
			peers_treeview.ShowNow();
			preferences_dialog.ShowNow();
		}

		public void peertreeviewrow_activated(object sender, EventArgs e){
			TreeModel tm;
			TreeIter iter;
			peers_treeview.Selection.GetSelected(out tm, out iter);
			string strPeer2rm = (string)mTreeStore.GetValue(iter,0);
			if(CConMan.PKnownNodes.Contains(strPeer2rm)){
				INode oNode = (INode)CConMan.PKnownNodes[strPeer2rm];
				ClNodeInfoDialog infoDia = new ClNodeInfoDialog(preferences_dialog,
																oNode,
																moPL,
																new DPeerListUpdateMethod(update_peers));
			}
		}

		public void add_peer(object sender, EventArgs e)
		{
			ClAddNodeDialog addDia = new ClAddNodeDialog(preferences_dialog,
														 moPL,
														 new DPeerListUpdateMethod(update_peers));
		}
		
		protected void close_preferences(object sender, EventArgs e)
		{
			if(nick_entry!=null && nick_entry.Text!=null){
				try{
					string strNick = nick_entry.Text;
					if(strNick!=CConMan.PLocalNickName){
						if (!File.Exists(strNick+"-cert.der")
							|| 
							!File.Exists(strNick+"-key.pem.der")){
							ClMsgDias.Err("Key or cert file not present in current directory.",
										  preferences_dialog);
						}
						else{
							CConMan.PLocalNickName = strNick;
							CConMan.PLocalX509Cert = Load(strNick+"-cert.der");
							ASN1 privateKey = new ASN1(Load(strNick+"-key.pem.der"));
							RSA rsa = PKCS8.PrivateKeyInfo.DecodeRSA(privateKey[2].Value);
							CConMan.PLocalRSA = rsa;
						}			
					}
					int newport = Convert.ToInt32(port_entry.Text);
					if(CConMan.PLocalListenPort!=newport){
						CConMan.PLocalListenPort=newport;
					}
				}
				catch(Exception errmsg){
					ClMsgDias.Err("Error:\n"+errmsg,preferences_dialog);
				}
				}
			preferences_dialog.Hide();	
		}

		private byte[] Load (string filename)
		{
			byte[] data = null;
			using (FileStream fs = File.OpenRead (filename)) {
				data = new byte [fs.Length];
				fs.Read (data, 0, data.Length);
			}
			return data;
		}
	}
}
