/* CSerializer - static class to de/serialize the XML classes
 * Also takes care of signing/validating
 * 
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 21.06.2004
 * Time: 21:49
 *
 */
 
 using System;
 using System.Xml.Serialization;
 using System.IO;
 using System.Collections;
 using System.Text.RegularExpressions;
 using System.Text;
 using System.Security.Cryptography;
 using Mono.Security.X509;
 
 using P2Pc.util;
 using P2Pc.net;
 
 namespace P2Pc.io{
 	
 	
 	/*
 	public interface IXml
 	{
 		//M:
 			
 			chatmessage[] parse(byte[] pTheMsg);
 		byte[] build(chatmessage pMsg);
 		
 	}
 	
 	*/
 	
 	public class CSerializer
 	{
 		#region Fields
 		private static XmlSerializer FSerializer;
 		private static MD5 Fmd5;
 		
 		private static UTF8Encoding FUtf8Enc;
 		private static ArrayList FQMessagesWaiting4Cert;
 		
 		private static X509Certificate _root = null;
 		private static X509Chain FCertChain;
 		
 		
 		#endregion
 		
 		
 		#region Constructors
 		static CSerializer(){
 			
 			FQMessagesWaiting4Cert = ArrayList.Synchronized(new ArrayList());
 			FSerializer = new XmlSerializer(typeof(chatmessage));
 			Fmd5 = new MD5CryptoServiceProvider();
 			
 			FUtf8Enc = new UTF8Encoding();
 			
 			//build the chain..
 			FCertChain = new X509Chain();
 			
 			try{
 				foreach(X509Certificate c in FCertChain.TrustAnchors)
 					if (c.IssuerName.IndexOf("pvsadm@ibr.cs.tu-bs.de") > 0)
 							_root=c;
 			}catch (Exception){_root=null; }
 			
 			if (_root == null){
 				Logger.Log("Serializer",3,"ca.cer not in root storage, creating...");
 				//hmm, pvsadmincert not in root storage.. copy it ;)
 				string path = Path.Combine (
 				                            Environment.GetFolderPath (Environment.SpecialFolder.ApplicationData),
 				                            ".mono");
 				path = Path.Combine (path, "certs");
 				path = Path.Combine (path, "Trust");
 				if (!File.Exists("ca.cer")){
 					Logger.Log("Serializer",3,"ca.cer not found, creating it from cacert.pem");
 					//no.cer file exists, have 2 create it out of cacert.pem
 					
 					if (!File.Exists("cacert.pem")) throw new ApplicationException("CA Certificate could not be found,please make sure at least cacert.pem is in the search path!!");
 					
 					StreamReader sr = new StreamReader("cacert.pem");
 					string data = sr.ReadToEnd();
 					if (data.StartsWith("------")){
 						//--------BEGIN CERT --- still there -> cut it
 						Match mt = Regex.Match(data,"-*BEGIN CERTIFICATE-*\n(.*)\n-*END CERTIFICATE-*",RegexOptions.Singleline);
 						data = mt.Groups[1].ToString();
 					}
 					
 					using (FileStream fs = File.Create ("ca.cer")) {
 						byte[] b = Convert.FromBase64String(data);
 						fs.Write(b,0,b.Length);
 						fs.Flush();
 					}
 					
 				}
 				Logger.Log("Serializer",3,"Trying to copy ca.cer to /trusts/ storage");
 				path = Path.Combine (path, "ca.cer");
 				try{
 					File.Copy("ca.cer",path);
 				}catch(Exception e){
 					Logger.Log("Serializer",2,"Copying of ca.cer failed:"+e);
 				}
 				
 				try{
 					foreach(X509Certificate c in FCertChain.TrustAnchors)
 						if (c.IssuerName.IndexOf("pvsadm@ibr.cs.tu-bs.de") > 0)
 							_root=c;
 				}catch (Exception){ _root=null; }
 				
 				if (_root==null){
 					Logger.Log("Serializer",2,"pvsadm cert still not in root storage, using ca.cer!");
 					//throw new ApplicationException("Critical: pvsadm-root cert not found!");
					//loading from ca.cer
					byte[] data = null;
					using (FileStream fs = File.OpenRead ("ca.cer")) {
						data = new byte [fs.Length];
						fs.Read (data, 0, data.Length);
					}
					_root = new X509Certificate(data); 
			
					
 				}
 				
 			}else Logger.Log("Serializer",4,"ca.cer (pvsadmin) found in root storage ;)");
 			//now we should have the ca.cer in the cain ;)
 		}
 		#endregion
 		
 		#region Properties
 		
 		
 		#endregion
 		
 		#region Methods
 		
 		#region parse
 		
 		/* static Method parse 
 		 * --------------------
 		 * Parses a byte array with UTF8 chars and tries to make a chatmessage class
 		 * out of it. Maybe more than one if it was a cert and other Msg were waiting
 		 * to be validated
 		 * 
 		 * in:	byte[] pTheMsg 	- the array
 		 * out:	chatmessage[]	- array of parsed chatmessages
 		*/
 		public static chatmessage[] parse(byte[] pTheMsg){
 			chatmessage cm = null;
 			
			//ugly hack.. for signtype="" _fucking other groups_
			string t = FUtf8Enc.GetString(pTheMsg);
 			Logger.Log("CMDeSerializer",5,"XMLString is:"+t);
			string t2 = t.Replace("signtype=\"\"","");
 			pTheMsg = FUtf8Enc.GetBytes(t2);
			
			
 			try{
 				cm = (chatmessage) FSerializer.Deserialize(new MemoryStream(pTheMsg));
 			}catch(Exception e){
 				Logger.Log("CMDeSerializer",2,"Deserialize error:"+e);
 				Logger.Log("CMDeSerializer",2,"XMLString was:"+FUtf8Enc.GetString(pTheMsg));
 				throw new EXMLParsingException("Desirializing failed:"+e);
 			}
 			
 			if (cm==null){
 				Logger.Log("CMDeSerializer",2,"Deserialize error");
 				throw new EXMLParsingException("Desirializing failed");
 			}
 			
 			//fill in the type in the NamedMsg... coz it does get overriden by derived classes (makes it all a little senseless ;)
 			if (cm.Item is ANamedMessage) ((ANamedMessage)cm.Item).CMType = cm.Item.GetType();
			else if (cm.Item is AMulticastMessage) ((AMulticastMessage)cm.Item).CMType = cm.Item.GetType();
			else if (cm.Item is AUnicastMessage) ((AUnicastMessage)cm.Item).CMType = cm.Item.GetType();
 			
 			IUser iu = null;
 			IUser iu2 = null;
 			chatmessage[] cm2=null;
 			ArrayList cma = new ArrayList();
 			X509Certificate cert;
 					
 			//check signature ... only if there...
			
 			if ((cm.Item is ASignedMessage) || (cm.Item is AMulticastMessage) || (cm.Item is AUnicastMessage) ){
				string sender = "";
 				if (cm.Item is ASignedMessage) sender=((ASignedMessage)cm.Item).sender;
				else if (cm.Item is AMulticastMessage) sender=((AMulticastMessage)cm.Item).sender;
				else if (cm.Item is AUnicastMessage) sender=((AUnicastMessage)cm.Item).sender;
 		
				
				if ( ( (cm.Item is CertificateMessage)
	 							&& (iu2=buildCert(cm)).PNick==sender) ){
 							cert=iu2.PCert;
							iu=iu2;
	 						Logger.Log("CMDeSerializer",4,"CertCM is self signed, adding",cm);
					}
				
				if ( ( sender==null || sender.Equals("") ) ){
					if (!(cm.Item is MessageMessage) ){
 					 	throw new EMessageValidationException("Signed CM without sender\n"+cm);
					}
				}else if ( (cm.signature==null || (cm.signature!=null&&cm.signature.Length<1) ) ){
					if ( (cm.Item is ChannelMessage) && ((ChannelMessage)cm.Item).closed){
					   throw new EMessageValidationException("unsigned CH CM 4 closed channel dropped for security reasons\n"+cm);
					}else Logger.Log("CMDeSerializer",2,"got unsigned signed CM, its a _SHOULD_,consider signing",cm);
				//else try to validate sig...
 			    } else {
	 				
	 				iu = CConMan.addToKnownUsers(sender);
 					//try to get the cert 4 this sender from the ConMan
 					bool bGCCM = false;
 					if (null == (cert=(iu.PCert)) ){
	 					//maybe its a getcert from one we dont have a cert from->ends in deadlock..
						if (cm.Item is GetCertificateMessage) bGCCM=true;
						//maybe its a CertCM itself? selfsigned is not very good.. but what the...
 						else if (!CConMan.FbOnlyNode){
 							Logger.Log("CMDeSerializer",2,"Cert not found, adding to queue",cm);
 							try{
								CConMan.sendGetCertificateMsg(iu);
								FQMessagesWaiting4Cert.Add(new WaitingQueueMsg(cm,pTheMsg));
								return null;
							}catch(Exception e){
								Logger.Log("CMDeSerializer",4,"sendGCM failed, accepting anyway..",e);
								cm2 = returnWaiting(iu);	
								
								if (cm2!=null && cm2.Length > 0) cma.AddRange(cm2);
 								bGCCM=true;
	 						}
 						}else bGCCM=true;
 					}
				
 					
 					//only check if not GetCertCM and not NodeOnly
 					if (!bGCCM && !ValidateSig(pTheMsg,cm,cert)){
 						throw new EMessageValidationException("Validation failed:\n"+cm);
 					}
				}
 			}
 			
 			//signature msg?
 			if (cm.Item is CertificateMessage)
 				cm2 = checkWaiting(iu);
 			
 			cma.Add(cm);
 			if (cm2!=null && cm2.Length > 0) cma.AddRange(cm2);
 			
 			return (chatmessage[])cma.ToArray(typeof(chatmessage));
 			
 		}
 		#endregion
 		
 		#region build
 		
 		/* static Method build
 		 * --------------------
 		 * Serializes a chatmessage class to its XML Representation
 		 * and takes care of signing...
 		 * 
 		 * in: 	chatmessage pMsg	- the chatmessage class to be serialized
 		 * in: 	bool pSign			- wether to sign the Msg
 		 * out:	byte[] 				- the array
 		 * 
 		*/
 		public static byte[] build(chatmessage pMsg, bool pSign){
 			
 			//build sig if > 5
 			if (pSign && ((pMsg.Item is ASignedMessage)  || (pMsg.Item is AMulticastMessage) || (pMsg.Item is AUnicastMessage)) && !(pMsg.Item is ObscureMessage) ){
 				
				if (CConMan.FbOnlyNode){
 					Logger.Log("CMSerializer",2,"tried to send new signed msg without local user",pMsg);
 					throw new EXMLParsingException("Trying to send new signed msg without local user");
 				}
 				
 				
				if (pMsg.Item is ASignedMessage) ((ASignedMessage)pMsg.Item).sender = CConMan.PLocalNickName;
				else if (pMsg.Item is AMulticastMessage) ((AMulticastMessage)pMsg.Item).sender = CConMan.PLocalNickName;
				else if (pMsg.Item is AUnicastMessage) ((AUnicastMessage)pMsg.Item).sender = CConMan.PLocalNickName;
				 //we need to serialize it first, to get a string
 				StringWriter sw = new StringWriter();
 				try{
 					FSerializer.Serialize(sw,pMsg);
 				}catch(Exception e){
 					Logger.Log("CMSerializer",2,"Serialize error",pMsg);
 					throw new EXMLParsingException("Parse Error:"+e);
 				}
 				
 				StringBuilder sb = new StringBuilder();
 				//match a 1337 regex against the XML 2 get all text between the tags
 				MatchCollection mt = Regex.Matches(sw.ToString(),"<([^/>]+)>([^<]+)</[^>]+>");
 				foreach (Match m in mt){
 					//$1 is a matched () pair aka. the string inside the <tags>, so add it to our string if not receiver element
 					if (!m.Groups[1].Equals("receiver")) sb.Append(m.Groups[2]);
 				}
 				//build sig and fill in pmsg field
 				pMsg.signature = ((RSACryptoServiceProvider)CConMan.PLocalRSA).SignData(
 				                                                             FUtf8Enc.GetBytes(sb.ToString()),
 				                                                             Fmd5);
 				pMsg.signtype = SignType.MD5;
 				pMsg.signtypeSpecified = true;
 				
 			}else pMsg.signtypeSpecified = false;
 			
 			
 			//use SWriter(memorystream) coz we want 2 get a UTF8 byte[]
 			MemoryStream ms = new MemoryStream();
			StreamWriter stw = new StreamWriter(ms,FUtf8Enc);
 			try{
 				FSerializer.Serialize(stw,pMsg);
 			}catch(Exception e){
 				Logger.Log("CMSerializer",2,"Serialize error",pMsg);
 				throw new EXMLParsingException("Serialize Error:"+e);
 			}
			Logger.Log("CMSerializer",5,"Serialized to XMLString:"+FUtf8Enc.GetString(ms.ToArray()));
				
			return ms.ToArray();
 			
 		}
 		#endregion
 		
 		#endregion
 		
 		#region privates
 		/* static Method buildCert
 		 * --------------------
 		 * tries to build a x509 Cert. out of a Cert Msg.
 		 * relies on CreateCert for the dirty stuff
 		 * also checks the trust status.. (but _only_ with the root CA , no chain - didnt work :(
 		 * 
 		 * in: 	chatmessage cmsg	- the chatmessage to be taken as a CertMsg
 		 * out:	IUser 				- the User to which the Cert was added
 		 * 
 		*/
 		private static IUser buildCert(chatmessage cmsg){
 			Logger.Log("X509Creator",5,"building x509Cert");
 			byte[] data = (byte[]) ((CertificateMessage)cmsg.Item).certificate.data.Clone();
 			X509Certificate cert = 	null;
 			
 			cert=CreateCert(data);
 			
 			if (cert == null){
 				Logger.Log("X509Creator",2,"Creation of x509Cert failed",cmsg);
 				return null;
 			}
 			Logger.Log("X509Creator",5,"x509Cert created, Subject:"+cert.SubjectName);
 			Logger.Log("X509Creator",5,"Isuer:"+cert.IssuerName);
 			
 			//we only have to check if the cert is trusted...
 			if (!isTrusted(cert)){
 				Logger.Log("X509Creator",1,"x509Cert not trusted",cmsg);
 				throw new ECertificateException("Certificate not trusted->Msg:"+cmsg);
 			}
 			
 			Logger.Log("X509Creator",5,"trust status ok");
 			
 			//get IUser... (match regex against full subject name 2 get E=.. part
 			Match mt = Regex.Match(cert.SubjectName,"E=(\\S+)",0);
 			string strUser = mt.Groups[1].ToString();
 			Logger.Log("X509Creator",4,"build cert for user "+strUser);
 			
 			IUser iu = null;
 			if (null == (iu = (IUser)CConMan.PKnownUsers[strUser]) )
 				iu = CConMan.addToKnownUsers(strUser);
 			// ...and add the cert
 			iu.PCert = cert;
 			iu.PbGCMPending = false;
 			Logger.Log("X509Creator",4,"build cert and added to User "+iu);
 			return iu;
 		}
 		
		/* static Method returnWaiting
 		 * --------------------
 		 * returns all Msgs which are waiting for this User in the Queue for Validation
 		 * 
 		 * in: 	IUser pUser			- the User to be checked for waiting Msgs
 		 * out:	chatmessage[]		- the waiting Msgs for the User
 		 * 
 		*/
 		private static chatmessage[] returnWaiting(IUser pUser){
 			if (FQMessagesWaiting4Cert.Count==0) return null;
 			ArrayList cma = new ArrayList();
 			
 			//check if there are waiting msgs to be validated
 			foreach (WaitingQueueMsg wqm in FQMessagesWaiting4Cert){
 				string ssender="";
				if (wqm.cm.Item is ASignedMessage) ssender = ((ASignedMessage)wqm.cm.Item).sender;
				else if (wqm.cm.Item is AMulticastMessage) ssender = ((AMulticastMessage)wqm.cm.Item).sender;
				else if (wqm.cm.Item is AUnicastMessage) ssender = ((AUnicastMessage)wqm.cm.Item).sender;
				
				if (pUser.PNick.Equals( (ssender ) ) ){
 					//msg is from certs User->	validate it
 					cma.Add(wqm.cm);
 				}
 			}
 			//now we have to rem the things we just added from the ArL
 			foreach (chatmessage cm in cma)
 				FQMessagesWaiting4Cert.Remove(cm);
 			
 			return (chatmessage[])cma.ToArray(typeof(chatmessage));
 		}
		
 		/* static Method checkWaiting
 		 * --------------------
 		 * checks all Msgs which are waiting for this User in the Queue for Validation
 		 * 
 		 * in: 	IUser pUser			- the User to be checked for Msgs
 		 * out:	chatmessage[]		- the waiting Msgs for the User
 		 * 
 		*/
 		private static chatmessage[] checkWaiting(IUser pUser){
 			
 			if (FQMessagesWaiting4Cert.Count==0) return null;
 			ArrayList cma = new ArrayList();
 				
 			//check if there are waiting msgs to be validated
 			foreach (WaitingQueueMsg wqm in FQMessagesWaiting4Cert){
 				string ssender="";
				if (wqm.cm.Item is ASignedMessage) ssender = ((ASignedMessage)wqm.cm.Item).sender;
				else if (wqm.cm.Item is AMulticastMessage) ssender = ((AMulticastMessage)wqm.cm.Item).sender;
				else if (wqm.cm.Item is AUnicastMessage) ssender = ((AUnicastMessage)wqm.cm.Item).sender;
				if (pUser.PNick.Equals( ssender ) ) {
 					//msg is from certs User->	validate it
 					if (ValidateSig(wqm.bt,wqm.cm,pUser.PCert)) cma.Add(wqm.cm);
 				}
 			}
 			//now we have to rem the things we just added from the ArL
 			foreach (chatmessage cm in cma)
 				FQMessagesWaiting4Cert.Remove(cm);
 			
 			return (chatmessage[])cma.ToArray(typeof(chatmessage));
 			
 		}
 		/* static Method CreateCert
 		 * --------------------
 		 * tries to build a x509 Cert. out of a a byte array
 		 * checks wether its a string or b64 encoded or nothing (then it fails)
 		 *
 		 * in: 	byte[] pData		- the Data
 		 * out:	X509Certificate		- the Cert
 		 * 
 		*/
 		private static X509Certificate CreateCert(byte[] pData){
 			string bytes="";
 			foreach (byte b in pData) bytes+=b.ToString();
 			
 			string certstr;
 			X509Certificate cert;
 			Logger.Log("X509Parser",5,"trying to create x509Cert from raw:"+bytes);
 			
 			try{
 				cert = new X509Certificate(pData);
 			}catch (Exception e){
 				Logger.Log("X509Parser",1,"creation of x509Cert failed",e);
 				cert = null;
 			}
 			if (cert != null){
 				Logger.Log("X509Parser",4,"Cert created, returning");
 				return cert;
 			}
 			
 			Logger.Log("X509Parser",5,"trying to create x509Cert from string");
 			try{
 				//if its just a string which was base64 encoded, it should be UTF8 encoded->decode
 				certstr = FUtf8Enc.GetString(pData);
 			}catch (Exception) {
 				//okey, so it wasnt
 				certstr = null;
 			}
 			if (certstr != null){
 				//ok, now match the (again) base64enc ASN1 cert part (betw. BEGIN / END CERT if necc.)
 				if (certstr.IndexOf("BEGIN CERTIFICATE") > 0){
 					Match mt = Regex.Match(certstr,"-*BEGIN CERTIFICATE-*\n(.*)\n-*END CERTIFICATE-*",RegexOptions.Singleline);
 					certstr = mt.Groups[1].ToString();
 				}
 				try{
 					pData = Convert.FromBase64String(certstr);
 					//now we should have only the binary ASN1 data...
 					cert = new X509Certificate(pData);
 				}catch(Exception e){
 					Logger.Log("X509Parser",1,"creation of x509Cert failed",e);
 					cert = null;
 				}
 			}
 			if (cert==null) Logger.Log("X509Parser",2,"Creation of x509Cert failed!");
 			return cert;
 			
 		}
 		
 		/* static Method ValidateSig
 		 * --------------------
 		 * checks a Signature
 		 * 
 		 * in: 	byte[] pMsg				- the UTF8 encoded array to be checked
 		 * in:  chatmessage pcm			- the Chatmessage, because the sig is in there
 		 * in:  X509Certificate pCert	- the Cert to be used for validation
 		 * out:	bool					- if its validated
 		 * 
 		*/
 		private static bool ValidateSig(byte[] pMsg, chatmessage pcm, X509Certificate pCert){
 			
 			StringBuilder sb = new StringBuilder();
 			//match a 1337 regex against the XML 2 get all text between the tags
 			MatchCollection mt = Regex.Matches(FUtf8Enc.GetString(pMsg),"<[^/>]+>([^<]+)</[^>]+>");
 			foreach (Match m in mt){
 				//$1 is a matched () pair aka. the string inside the <tags>, so add it to our string
 				sb.Append(m.Groups[1]);
 			}
 			
 			//get RSA from Cert and check sig
 			bool ret = 	((RSACryptoServiceProvider)pCert.RSA).VerifyData(FUtf8Enc.GetBytes(sb.ToString()),	//the buffer to hash (in UTF8)
 			                                                             Fmd5,				//hash alg.
 			                                                             pcm.signature		//the sig 2 check
 			                                                             );
 			if (!ret) Logger.Log("SigValidator",2,"Validation failed",pcm);
 			return ret;
 		}
 		
 		/* static Method isTrusted
 		 * --------------------
 		 * returns all Msgs which are waiting for this User in the Queue for Validation
 		 * 
 		 * in: 	X509Certificate pCert	- the Cert
 		 * out:	bool					- the trust status
 		 * 
 		*/
 		private static bool isTrusted(X509Certificate pCert){
 		    return ( _root==null || ( (pCert.IssuerName == _root.SubjectName)
 		            && (pCert.VerifySignature(_root.RSA)) ));
		}
 		
 		#endregion
 		
 		#region private queue entry class
 		
 		private sealed class WaitingQueueMsg{
 			public chatmessage cm;
 			public byte[] bt;
 			public WaitingQueueMsg(chatmessage pCm,byte[] pTheMsg){
 				cm = pCm;
 				bt = pTheMsg;
 			}
 		}
 		#endregion
 		
 		
 	}
 	
 	
 }
