/* CEncryption - static class that en- and decrypts things
 * has methods for Asym(RSA) and Symm. (DES) Encryption
 * 
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 28.06.2004
 * Time: 12:46
 *
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
 using System;
 using System.Security.Cryptography;
 using Mono.Security.X509;
 using Mono.Security.Cryptography;
 using System.IO;
 
 using P2Pc.util;
 using P2Pc.net;
 using System.Text;
 
 namespace P2Pc.io{
 	
 	
 	//Interface cannot have static members, we have to comment it out
 	/*
 	public interface IEnc
 	{
 		//M:
 			
 			string SymEncrypt(string msg, SymmetricAlgorithm sa);   //needs to be a string, so we have to b64 encode it
 			string SymDecrypt(string msg, SymmetricAlgorithm sa);
 		//for KEY/OBSCURE msgs (wants b64 binary)
 		byte[] AsymEncrypt(byte[] msg, AsymmetricAlgorithm aa);
 		byte[] AsymDecrypt(byte[] msg, AsymmetricAlgorithm aa);
 		
 	}
 	*/
 	
 	public class CEncryption
 	{
 		#region Fields
 		private static UTF8Encoding FUtf8Enc;
 		#endregion
 		
 		
 		#region Constructors
 		static CEncryption ()
 		{
 			FUtf8Enc = new UTF8Encoding();
 			
 		}
 		#endregion
 		
 		
 		#region Methods
 		//----------------------------------------------------------------------------
 		//following SymEncryption funcs are only used for closed channels, and since MSG msg uses string
 		//for its text element we can savely operate with strings (only have to b64 en/decode them ;)
 		
 		/* static Method SymEncrypt 
 		 * --------------------
 		 * Encrypt a string symm. with DES.
 		 * Note: needs to be mult of 64.. so padd it
 		 * 
 		 * in:	string msg	 	- the msg to be enc.
 		 * in:	DES	sa		 	- the DES-CSP to be used
 		 * out:	string			- the encrypted string
 		*/
 		public static string SymEncrypt(string msg, DES sa){
 			Logger.Log("SymEncrypt",5,"encrypting \""+msg+"\"");
 			
 			//needs to be UTF8 encoded byte[], so encode it
 			byte[] data = FUtf8Enc.GetBytes(msg);
 			
 			//look if data/64!=0 byte.. if yes, have to pad it with 000
 			if (data.Length%64 > 0){
 				Logger.Log("SymEncrypt",5,"msg not mult. of 64("+data.Length+"), padding with 0 to c*64bytes");
 				byte[] data2 = new byte[ (data.Length/64+1) * 64];
 				Buffer.BlockCopy(data,  //src
 				                 0,				//src-start
 				                 data2,		//dst
 				                 0,			//dst-start
 				                 data.Length);
 				data=data2;
 			}
 			sa.GenerateIV();
 			MemoryStream ms = new MemoryStream();
 			ms.SetLength(0);
 			
 			try{
 				CryptoStream encStream = new CryptoStream(ms, sa.CreateEncryptor(), CryptoStreamMode.Write);
 				encStream.Write(data,0,data.Length);
 				encStream.Flush();
 				ms.Flush();
 			}catch (Exception e){
 				Logger.Log("SymEncrypt",5,"error encrypting",e);
				throw new EEncryptingException("Encrypting failed:"+e);
 			}
 			Logger.Log("SymEncrypt",5,"encrypting "+data.Length+" bytes of data");
 			
			string bytes="";
 			if (Logger.PLevel>4) foreach (byte b in ms.ToArray()) bytes+=b.ToString();
 			Logger.Log("SymEncrypt",5,"encrypted to "+ms.Length+" bytes of data:"+bytes);
 			
 			//now we have the binary data, to put it back into a string, we need to b64enc it...
 			return Convert.ToBase64String(ms.ToArray());
 			
 		}
 		
 		/* static Method SymDecrypt 
 		 * --------------------
 		 * Decrypts a string symm. with DES.
 		 * 
 		 * in:	string msg	 	- the msg to be dec.
 		 * in:	DES	sa		 	- the DES-CSP to be used
 		 * out:	string			- the decrypted string
 		*/
 		public static string SymDecrypt(string msg, DES sa){
 			Logger.Log("SymDecrypt",5,"decrypting \""+msg+"\"");
 			//should be base64 encoded binary data -> decode
 			byte[] data=null;
 			byte[] data2=null;
 			try{
 				data = Convert.FromBase64String(msg);
 			}catch(Exception e){
 				//maybe not...
 				Logger.Log("SymDecrypt",2,"b64 decode error",e);
 				throw new EDecryptingException("Decoding base64 failed->"+e+"\nMsg:"+msg);
 			}
 			string bytes="";
 			if (Logger.PLevel>4) foreach (byte b in data) bytes+=b.ToString();
 			Logger.Log("SymDecrypt",5,"decrypting "+data.Length+" bytes of data:"+bytes);
 			
			//now we have the binary->decrypt
 			MemoryStream ms = new MemoryStream();
 			ms.SetLength(0);
			string ret="";
 			try{
 				CryptoStream encStream = new CryptoStream(ms, sa.CreateDecryptor(), CryptoStreamMode.Write);
 				encStream.Write(data,0,data.Length);
 				encStream.Flush();
 			
 				ms.Flush();
 				Logger.Log("SymDecrypt",5,"decrypted to "+ms.Length+" bytes of data,removing trailing zeros");
 				data2=RemoveTrailingZeroes(ms.ToArray());
 				ret = FUtf8Enc.GetString(data2);
				Logger.Log("SymDecrypt",5,"zeros removed, final bytes("+data2.Length+"), final string:"+ret);
 			}catch (Exception e){
 				Logger.Log("SymDecrypt",5,"error decrypting",e);
				throw new EDecryptingException("Decrypting failed:"+e);
 			}		
			return ret;
 			
 		}
 		
 		//---------------------------------------------------------------------------
 		//4 KEY/OBSCURE msgs
 		
 		
 		/* static Method AsymEncrypt
 		 * --------------------
 		 * Encrypts a string asymm. with RSA.
 		 * Note: needs to be mult of 117.. so padd it
 		 * 
 		 * in:	byte[] msg	 	- the msg to be enc.
 		 * in:	RSAManged aa 	- the ESA-CSP to be used
 		 * out:	byte[]			- the encrypted array
 		*/
 		public static byte[] AsymEncrypt(byte[] msg, RSAManaged aa){
 			string bytes="";
 			if (Logger.PLevel>4) foreach (byte b in msg) bytes+=b.ToString();
 			Logger.Log("AsymEncrypt",5,"encrypting("+msg.Length+") "+bytes);
 			
 			//msg > 117 byte?
 			RNGCryptoServiceProvider rngcsp = new RNGCryptoServiceProvider();
 			int i = 0;int thisBlock=0;
 			byte[] tempBuffer;
 			byte[] outBuffer = new byte[ ((msg.Length/117)+1) * 128];
 			
 			while ( msg.Length > (i*117) ){
 				thisBlock = Math.Min(msg.Length-(i*117),117);
 				tempBuffer = new byte[thisBlock];
 				//copy at most 117bytes to buffer and encrypt
 				Buffer.BlockCopy(msg,i*117,tempBuffer,0,thisBlock);
 				Buffer.BlockCopy(PKCS1.Encrypt_v15(aa, rngcsp, tempBuffer),  //src
 				                 0,				//src-start
 				                 outBuffer,		//dst
 				                 i*128,			//dst-start
 				                 128			//length
 				                 );
 				//and count up i
 				i++;
 			}
 			
 			Logger.Log("AsymEncrypt",5,"encrypted to "+outBuffer.Length+" bytes");
 			return outBuffer;
 			//return ((RSACryptoServiceProvider)aa).Encrypt(msg,false);
 		}
 		
 		/* static Method AsymDecrypt 
 		 * --------------------
 		 * Decrypts a string asymm. with RSA.
 		 * 
 		 * in:	byte[] msg	 	- the msg to be dec.
 		 * in:	RSAManaged aa	- the RSA-CSP to be used
 		 * out:	byte[]			- the decrypted array
 		*/
 		public static byte[] AsymDecrypt(byte[] msg, RSAManaged aa){
 			//split in 128byte chunks and decrypt
 			int i = 0;
 			byte[] tempBuffer = new byte[128];
 			byte[] outBuffer = new byte[ (msg.Length/128) * 117];
 			byte[] dec=null;
 			Logger.Log("AsymDecrypt",5,"decrypting "+msg.Length+" bytes");
 			while ( msg.Length > (i*128) ){
 				dec = null;
 				//Logger.Log("AsymEncrypt",5,"decrypting "+i+"th 128 bytes");
 				//copy 128 bytes to buffer and decrypt
 				Buffer.BlockCopy(msg,i*128,tempBuffer,0, 128);
 				dec = PKCS1.Decrypt_v15(aa,tempBuffer);
 				Buffer.BlockCopy(dec,  //src
 				                 0,				//src-start
 				                 outBuffer,		//dst
 				                 i*117,			//dst-start
 				                 dec.Length		//length
 				                 );
 				//and count up i
 				i++;
 			}
 			//Logger.Log("AsymDecrypt",5,"decrypted "+outBuffer.Length+" bytes,with zeros");
 			
 			string bytes="";
 			if (Logger.PLevel>4) foreach (byte b in outBuffer) bytes+=b.ToString();
 			Logger.Log("AsymDecrypt",5,"decrypted("+outBuffer.Length+") "+bytes+" with zeros");
 			byte[] data2=RemoveTrailingZeroes(outBuffer);
 			
			bytes="";
			if (Logger.PLevel>4) foreach (byte b in data2) bytes+=b.ToString();
 			Logger.Log("AsymDecrypt",5,"zeros removed, final bytes("+data2.Length+"):"+bytes);
 			
			return data2;
 			
 		}
 		#endregion
 		
 		/* static Method RemoveTrailingZeroes 
 		 * --------------------
 		 * Removes the trailing zeros which might have been added
 		 * 
 		 * in:	byte[] input 	- the input
 		 * out:	byte[]			- array without trailing 0's
 		*/
 		private static byte[] RemoveTrailingZeroes(byte[] input){
 			int i=input.Length-1;
 			while (input[i]==0x00) i--;
 			//now we have last not 00 val
 			byte[] outBuffer = new byte[i+1];
 			Buffer.BlockCopy(input,  //src
 			                 0,				//src-start
 			                 outBuffer,		//dst
 			                 0,			//dst-start
 			                 i+1 );			//length
 			                 return outBuffer;
 			
 		}
 		
 	}
 }
