/* CChannel - class that represents a channel
 * taked care of all Channel related stuff (event notification, Msg Building..)
 *
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 05.07.2004
 * Time: 13:53
 *
 */
 using System;
 using System.Collections;
 using System.Net.Sockets;
 using System.Security.Cryptography;
 using Mono.Security.Cryptography;

 using P2Pc.io;
 using P2Pc.util;

 namespace P2Pc.net{

 	public class CChannel : IChannel{
 		#region Fields

 		private bool		FbClosed;
 		private string		FName;
 		private string		FTopic;
 		private DES			FSymAlgo;
 		private ArrayList	FUsers;
 		private DateTime	FLastChannelMsg;

 		#endregion

 		#region Events
 		public event DChannelEventHandler EvChannelEvent;
 		#endregion

 		#region Constructors
 		public CChannel(string pName){
 			FName = pName;
 			FUsers = ArrayList.Synchronized(new ArrayList());
 			FLastChannelMsg = DateTime.Now;

 		}

 		public CChannel(string pName,bool closed) : this(pName){
 			FbClosed=closed;

 			if (FbClosed){		//have to create a key
 			FSymAlgo = new DESCryptoServiceProvider();
 				FSymAlgo.Mode = CipherMode.CBC;
 				FSymAlgo.Padding = PaddingMode.PKCS7;
 				FSymAlgo.GenerateKey();
 				//add local user

 			}
 		}
 		public CChannel(string pName,bool closed,bool autoJoin) : this(pName,closed){
 			try{
 				if (autoJoin) FUsers.Add(CConMan.PLocalUser);
 			}catch (Exception){}
 		}
 		public CChannel(string pName,byte[] key) : this(pName){
 			FbClosed=true;
 			FSymAlgo = new DESCryptoServiceProvider();
 			FSymAlgo.Mode = CipherMode.CBC;
 			FSymAlgo.Padding = PaddingMode.PKCS7;
 			FSymAlgo.Key = key;

 		}
 		public CChannel(ChannelMessage cm){

 			if ( (FName = cm.channel) != ""){
 				FTopic = cm.description;
 				FbClosed = cm.closed;
 				FUsers = new ArrayList();
 				if (cm.member!=null) foreach (string mem in cm.member)
 					FUsers.Add(CConMan.addToKnownUsers(mem));

 				if (FbClosed){
 					FSymAlgo = new DESCryptoServiceProvider();
 					FSymAlgo.Mode = CipherMode.CBC;
 					FSymAlgo.Padding = PaddingMode.PKCS7;
 				}
 				FLastChannelMsg = DateTime.Now;
 			}else throw new ENetException("void channel is omnipresent - cannot be created!");
 		}

 		#endregion

 		#region Methods
 		public void		sendMessage(string msg){
 			if (FbClosed && !PbLocalUserMember){
 				Logger.Log("Channel:"+FName,1,"tried to post msg to closed channel without being member");
 				throw new ENetException("tried to post msg to closed channel "+FName+" without being member");
 			}else{
 				chatmessage  cm = new chatmessage(typeof(MessageMessage),32,false);
 				MessageMessage mm = (MessageMessage)cm.Item;
 				if (FName!="") mm.channel = this.FName;
 				mm.text = new MessageMessageText();

 				if (FbClosed){
 					mm.text.Value = CEncryption.SymEncrypt(msg, FSymAlgo);
 					mm.text.iv = FSymAlgo.IV;
 				}else mm.text.Value = msg;

 				if (FName==""){
 					cm.flood = true;
 					CConMan.floodCM(cm,null);
 				}else{
 					ArrayList ToAl = new ArrayList(FUsers);
 					CConMan.SendToUsers(ToAl,cm,0,true);
 				}
 			}
 			//and raise event
 			onChannelEvent(ChannelEventType.Message,CConMan.PLocalUser,msg);
 			FLastChannelMsg = DateTime.Now;

 		}


 		public void		acceptUser(IUser user){
 			FLastChannelMsg = DateTime.Now;
 			if (FName=="") return;
 			if (FbClosed){
 				//we have to build new ChannelMsg with this user in it ;)
 				if (!FUsers.Contains(user)) FUsers.Add(user);
 				chatmessage cm = buildChannelCM();
 				CConMan.floodCM(cm,null);
 				onChannelEvent(ChannelEventType.UserJoined,user,"");
 			}
 		}

 		public void		join(){
 			FLastChannelMsg = DateTime.Now;
 			if (FName=="") return;

 			chatmessage  cm = new chatmessage(typeof(JoinMessage),32,true);
 			((JoinMessage)cm.Item).channel = this.FName;
 			CConMan.floodCM(cm,null);
 			if (!FbClosed){
 				if (!FUsers.Contains(CConMan.PLocalUser)){
 					FUsers.Add(CConMan.PLocalUser);
 					onChannelEvent(ChannelEventType.UserJoined,CConMan.PLocalUser,"");
 				}
 			}


 		}

 		public void		leave(){
 			FLastChannelMsg = DateTime.Now;
 			if (FName=="") return;
 			chatmessage  cm = new chatmessage(typeof(LeaveMessage),32,true);
 			((LeaveMessage)cm.Item).channel = this.FName;
 			CConMan.floodCM(cm,null);
 			if (FUsers.Contains(CConMan.PLocalUser)){
 				FUsers.Remove(CConMan.PLocalUser);
 				onChannelEvent(ChannelEventType.UserLeft,CConMan.PLocalUser,"");
 			}


 		}

 		public chatmessage buildChannelCM(){
 			if (FName=="") return null;

 			chatmessage cm = new chatmessage(typeof(ChannelMessage),32,true);
 			ChannelMessage ChanM = (ChannelMessage)cm.Item;
 			ChanM.channel = FName;
 			ChanM.description = FTopic;
 			ChanM.closed = FbClosed;
 			ArrayList UserA = new ArrayList();
 			try{
 				foreach (IUser iu in FUsers) UserA.Add(iu.PNick);
 			}catch (Exception){}

 			ChanM.member = (string[])UserA.ToArray(typeof(string));
 			FLastChannelMsg = DateTime.Now;
 			return cm;
 		}

 		#region Process Messages
 		public bool processCM(ChannelMessage pcm, IAbstractNetInfo sender){
 			//if ""CH or sender is not member: dont process and dont reflood
 			if (pcm.channel=="") return false;
 			else if (!(FUsers.Contains(CConMan.addToKnownUsers(pcm.sender))) ) return true;

 			bool bLocalFound = false,ret=true;
 			CUser cu;
 			FLastChannelMsg = DateTime.Now;
 			//set the topic...
 			if (pcm.description!=null && pcm.description!=""){
 				if (FTopic==null || !FTopic.Equals(pcm.description)) onChannelEvent(ChannelEventType.TopicSet,CConMan.addToKnownUsers(pcm.sender),pcm.description);
 				FTopic = pcm.description;
 			}

 			//add new people
 			if (pcm.member!=null) foreach (string mem in pcm.member){
 				cu = CConMan.addToKnownUsers(mem);
 				if (cu == CConMan.PLocalUser){
 					bLocalFound=true;

 					//see if its a faulty msg with us still in it, and not hinder joining of closed CH
 					if (!PbLocalUserMember && !FbClosed){
 						leave();
 						ret= false;
 					}
 				}

 				if (!FUsers.Contains(cu)){		//new user in channel
 				FUsers.Add(cu);
 					//raise event
 					onChannelEvent(ChannelEventType.UserJoined,cu,"");

 					//see if we were accepted to a closed channel
 					if (FbClosed && cu == CConMan.PLocalUser){
 						//have to send getkey, to get the key
 						chatmessage cm = new chatmessage(typeof(GetKeyMessage),32,false);
 						((GetKeyMessage)cm.Item).channel = FName;
 						((GetKeyMessage)cm.Item).receiver = pcm.sender;
 						((CNode)sender).sendCM(cm);
 					}
 				}
 			}
 			//add us, if were not listed...
 			if (PbLocalUserMember && !bLocalFound){
 				join();
 				ret=false;
 			}

 			return ret;
 		}

 		public void processJM(JoinMessage jm, IAbstractNetInfo sender){
 			if (jm.channel=="") return;
 			string mem = ((ASignedMessage)jm).sender;
 			CUser cu = CConMan.addToKnownUsers(mem);

 			//check if is a closed channel.. then we have to ask
 			if (FbClosed)
 				onChannelEvent(ChannelEventType.UserWantsToJoin,cu,"");
 			else{		//just add the user
 				if (!FUsers.Contains(cu)) FUsers.Add(cu);
 				onChannelEvent(ChannelEventType.UserJoined,cu,"");

 			}
 			FLastChannelMsg = DateTime.Now;

 		}
 		public void processLM(LeaveMessage lm, IAbstractNetInfo sender){
 			if (lm.channel=="") return;
 			string mem = ((ASignedMessage)lm).sender;
 			CUser cu = CConMan.addToKnownUsers(mem);

 			//check if is a closed channel.. then it makes no sense
 			if (!FbClosed && FUsers.Contains(cu)){
 				FUsers.Remove(cu);
 				onChannelEvent(ChannelEventType.UserLeft,cu,"");
 			}
 			FLastChannelMsg = DateTime.Now;

 		}
 		public void processMM(MessageMessage mm, IAbstractNetInfo sender){
 			FLastChannelMsg = DateTime.Now;
 			if (!PbLocalUserMember) return;
 			//closed?
 			string msg="";
 			if (FbClosed){
 				try{
 					FSymAlgo.IV = mm.text.iv;
 					msg = CEncryption.SymDecrypt(mm.text.Value, FSymAlgo);
 				}catch (Exception e){
 					//no key?
 					if (FSymAlgo.Key==null || FSymAlgo.Key.Length==0){
 						chatmessage cm = new chatmessage(typeof(GetKeyMessage),32,false);
 						((GetKeyMessage)cm.Item).channel = FName;
 						((GetKeyMessage)cm.Item).receiver = mm.sender;
 						((CNode)sender).sendCM(cm);
 					}else
 						throw new ENetException("SymDecrypting of Msg failed:"+e);
 				}

 			}else msg = mm.text.Value;
 			CUser cu=null;
 			if (mm.sender!=null && mm.sender!="")
 				cu = (CUser)CConMan.PKnownUsers[mm.sender];
 			if (msg!="") onChannelEvent(ChannelEventType.Message,cu,msg);

 		}


 		#endregion

 		public chatmessage buildKM(GetKeyMessage gkm, CUser p4User){
 			if (!FbClosed || !PbLocalUserMember || !(FUsers.Contains(p4User)) ) return null;

 			chatmessage cm = new chatmessage(typeof(KeyMessage),32,false);

 			KeyMessage km = (KeyMessage)cm.Item;
 			km.receiver = gkm.sender;
 			km.channel = FName;
 			km.cipher = CipherType.DESCBC;
 			km.key = CEncryption.AsymEncrypt(FSymAlgo.Key,p4User.PRSAM);
 			FLastChannelMsg = DateTime.Now;

 			return cm;
 		}

 		public override string ToString(){
 			string us="";
 			if (Logger.PLevel > 4){
 				foreach(CUser cu in FUsers){
 					us += ((us!="") ? ",":"") + cu.PNick ;
 				}
 				us = ",{"+us+"}";
 			}
 			return "Channel{"+FName+","+FTopic+","+FbClosed+","+FUsers.Count+us+"}";
 		}

 		private void onChannelEvent(ChannelEventType cet,IUser iu,string msg){
 			ChannelEventArgs cea = new ChannelEventArgs(cet,iu,msg);
 			if (EvChannelEvent!=null) EvChannelEvent(this,cea);
 			Logger.Log("Channel:"+FName,3,cet+"-Event("+((iu!=null)?iu.PNick:"anon")+")");
 		}

 		#endregion


 		#region Properties
 		public bool PbLocalUserMember
 		{
 			get{
 				return ( FName=="" || FUsers.Contains(CConMan.PLocalUser)  );
 			}
 			set{
 				if (value) join();
 				else leave();
 			}
 		}

 		public bool PbClosed
 		{
 			get{
 				return FbClosed;
 			}
 		}
 		public string PName
 		{
 			get{
 				return FName;
 			}
 		}

 		public string PTopic
 		{
 			get{
 				return FTopic;
 			}
 			set{
 				FTopic = value;
 				CConMan.floodCM(buildChannelCM(),null);
 				onChannelEvent(ChannelEventType.TopicSet,CConMan.PLocalUser,FTopic);
 			}
 		}

 		public byte[] PSymKey
 		{
 			get{
 				return this.FSymAlgo.Key;
 			}
 			set{
 				FSymAlgo.Key = value;
 				Logger.Log("Channel:"+FName,3,"Key set");
 			}
 		}

 		public SymmetricAlgorithm PSymAlgo
 		{
 			get{
 				return FSymAlgo;
 			}
 		}

 		public IUser[] PUsers
 		{
 			get{
 				return (IUser[])FUsers.ToArray(typeof(IUser));
 			}
 		}

 		public DateTime PLastChannelMessage
 		{
 			get{
 				return FLastChannelMsg;
 			}
 		}
 		#endregion



 	}

}
