/* static class CConMan - Connection Manager
 * Handles all connection related stuff (high level Interface) 
 * Is static because we only need it once
 *
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 02.07.2004
 * Time: 16:19
 *
 */
 
 using System;
 using System.Net;
 using System.Collections;
 using System.Timers;
 using System.Security.Cryptography;
 using System.Net.Sockets;
 using System.IO;
 using Mono.Security.Cryptography;
 
 using P2Pc.io;
 using P2Pc.util;
 
 namespace P2Pc.net{
 	
 	public class CConMan
 	{
 		
 		#region Fields
 		
 		private static bool FbSendRandomAnonMsgs = false;
 		private static string FLocalNickName = null;
 		private static CUser FLocalUser;
 		private static IPEndPoint FLocalEndPoint;
 		private static int FLocalListenPort = 4321;
 		private static Socket FLocalListeningSocket;
 		private static RSA FLocalRSA;
 		private static byte[] FLocalX509Cert;
 		private static Hashtable FKnownChannels;
 		private static Hashtable FKnownNodes;
 		private static Hashtable FKnownUsers;
 		private static Timer 	FTimer;
 		private static byte FConnCount = 0;
 		public  static bool FbOnlyNode = true;
 		#endregion
 		
 		#region Events
 		public static event DNetInfoChangedEventHandler EvNetInfoChanged;
 		#endregion
 		
 		#region Constructors
 		
 		static CConMan(){
 			
 			//init vars
 			FKnownChannels = 	Hashtable.Synchronized(new Hashtable());
 			FKnownNodes = 		Hashtable.Synchronized(new Hashtable());
 			FKnownUsers = 		Hashtable.Synchronized(new Hashtable());
 			
 			//create the default channel
 			addToKnownChannels("");
 			
 			//start timer
 			FTimer = new Timer(10000);
 			FTimer.Elapsed+=new ElapsedEventHandler(OnTimerEvent);
 			FTimer.Enabled = true;
 			
 			//and start listening
 			//StartListening();
 			Logger.Log("ConMan",0,"initialized...");
 			
 		}
 		
 		
 		public static void Destroy(){
 			try{
 				FTimer.Enabled=false;
 				FTimer.Stop();
 				FTimer=null;
 				globalDisConnect();
 				//FLocalListeningSocket.Shutdown(SocketShutdown.Both);
 				FLocalListeningSocket.Close();
 				FLocalListeningSocket=null;
 				FKnownChannels.Clear();
 				FKnownNodes.Clear();
 				FKnownUsers.Clear();
 				FKnownChannels = null;
 				FKnownNodes = null;
 				FKnownUsers = null;
 				Logger.Log("ConMan",0,"Destroyed...");
 			}catch(Exception e){
 				Logger.Log("ConMan",3,"Destroying",e);
 			}
 		}
 		#endregion
 		
 		
 		#region Methods
 		
 		public static void SaveKnownNodes(Stream pStream){
 			System.Xml.Serialization.XmlSerializer xmls = new System.Xml.Serialization.XmlSerializer(typeof(PeersInfo));
 			
 			PeersInfo pi = new PeersInfo(FKnownNodes.Values);
 			xmls.Serialize(pStream,pi);
 			Logger.Log("ConMan",3,"saved known Nodes to file");
 			
 		}
 		
 		public static void LoadKnownNodes(Stream pStream){
 			System.Xml.Serialization.XmlSerializer xmls = new System.Xml.Serialization.XmlSerializer(typeof(PeersInfo));
 			PeersInfo pi = (PeersInfo)xmls.Deserialize(pStream);
 			foreach (NodeSerBase nsb in pi.peer){
 				if (nsb.Metric > -1){
 					if (!FKnownNodes.Contains(nsb.Address)){
 						int idx = nsb.Address.LastIndexOf(':');
 						int port = Int32.Parse(nsb.Address.Substring(idx+1));
 						string ip =  nsb.Address.Substring(0,idx);
 						IPAddress ipa=null;
 						try{
 							ipa = IPAddress.Parse(ip);
 						}catch(Exception){}
 						if (ipa!=null){
 							CNode cn = new CNode(new IPEndPoint(ipa,port),nsb.Metric);
 							FKnownNodes.Add(nsb.Address,cn);
 							cn.EvStateChanged+=new DNodeChangedEventHandler(NodeStateChanged);
 							onNetInfoChanged(cn,NetInfoEventType.node);
 						}
 					}
 				}
 			}
 			pi=null;
 			Logger.Log("ConMan",3,"loaded known Nodes from file");
 		}
 		
 		#region Process CM (main method)
 		
 		/* static Method processChatMessage
 		 * --------------------
 		 * Processes the ChatMessage given.
 		 * Takes care of _all_ stuff related to Msgs...
 		 * Knows all types, all Info and stuff.
 		 * 
 		 * in:	IAbstractNetInfo sender 	- the sender (Node) from where the Msg came
 		 * in:	chatmessage					- the ChatMessage to process
 		*/
 		public static void processCM(chatmessage pcm, IAbstractNetInfo sender){
 			//all msg types
 			//HelloMessage hm;
 			ChannelMessage chanm;
 			JoinMessage jm;
 			LeaveMessage lm;
 			//GetPeersMessage gpm;
 			PeersMessage pm;
 			GetCertificateMessage gcm;
 			CertificateMessage certm;
 			GetKeyMessage gkm;
 			KeyMessage km;
 			RoutingMessage rm;
 			MessageMessage mm;
 			ObscureMessage om;
 			
 			//some other often used vars
 			CNode CNSender = (CNode)sender;
 			CChannel cc;
 			CUser cu;
 			//and count down ttl
 			--pcm.ttl;
 			
 			Logger.Log("ConMan",3,"processing CM:",pcm);
 			//type of cm? note: switch only worx with integral types...
 			
 			#region Hello
 			if (pcm.Item is HelloMessage) {
 				//have to send CHANNEL CM for each known channel
 				if (!FbOnlyNode){
 					//get sender
 					try{
 						foreach (CChannel cc2 in FKnownChannels.Values)
 							if (cc2.PName != "")
 								CNSender.sendCM(cc2.buildChannelCM());
 					}catch (Exception e){
 						throw new ENetException("error while sending CHANNEL CM:"+e);
 					}
 					
 				}
 				//send ROUTING CM for this link
 				CNSender.sendCM(buildRoutingCM(CNSender));
 				
 			}
 			#endregion
 			#region Channel
 			else if (null != (chanm = pcm.Item as ChannelMessage) ){
 				bool bResend=true;
 				if (chanm.channel==null) chanm.channel="";
 				if (null!= ( (cc = FKnownChannels[chanm.channel] as CChannel) ) )
 					bResend = cc.processCM(chanm,sender);
 				else
 					//channel not known->create...
 					addToKnownChannels(chanm);
 				
 				//and resend to other links
 				if (pcm.ttl == 0) Logger.Log("ConMan",2,"ttl reached 0",pcm);
 				else if (bResend) floodCM(pcm,CNSender);
 				onNetInfoChanged(cc,NetInfoEventType.channel);
 			}
 			#endregion
 			#region Join
 			else if (null != (jm = pcm.Item as JoinMessage) ){
 				cc=addToKnownChannels(jm.channel);
 				cc.processJM(jm,CNSender);
 				//and resend to other links
 				if (pcm.ttl == 0) Logger.Log("ConMan",2,"ttl reached 0",pcm);
 				else floodCM(pcm,CNSender);
 				onNetInfoChanged(cc,NetInfoEventType.channel);
 			}
 			#endregion
 			#region Leave
 			else if (null != (lm = pcm.Item as LeaveMessage) ){
 				cc=addToKnownChannels(lm.channel);
 				cc.processLM(lm,sender);
 				//and resend to other links
 				if (pcm.ttl == 0) Logger.Log("ConMan",2,"ttl reached 0",pcm);
 				else floodCM(pcm,CNSender);
 				onNetInfoChanged(cc,NetInfoEventType.channel);
 			}
 			#endregion
 			#region GetPeers
 			else if (pcm.Item is GetPeersMessage) {
 				CNSender.sendCM(buildPeersCM(CNSender));
 			}
 			#endregion
 			#region Peers
 			else if (null != (pm = pcm.Item as PeersMessage) ){
 				//add all peers from the msg to knownNodes if not already known
 				if (pm.peer!=null) foreach(PeerAddressType pat in pm.peer){
 					IPEndPoint ep;
 					if (pat.ItemElementName == ItemChoiceType.ipaddress)
 						ep = new IPEndPoint(IPAddress.Parse(pat.Item),pat.port);
 					else{
 						//hmm, have to query for
 						IPHostEntry hI = Dns.GetHostByName(pat.Item);
 						ep = new IPEndPoint(hI.AddressList[0],pat.port);
 					}
 					//check if already known...
 					if (! FKnownNodes.Contains(ep.ToString()) )
 						addToKnownNodes(ep);
 					
 				}
 				
 			}
 			#endregion
 			#region GetCertificate
 			else if (null != (gcm = pcm.Item as GetCertificateMessage) ){
 				if (gcm.receiver!=null){
 					if (gcm.receiver.Equals(FLocalNickName)){
 						chatmessage cm = new chatmessage(typeof(CertificateMessage),32,false);
 							((CertificateMessage)cm.Item).receiver = new String[] {gcm.sender} ;
 						((CertificateMessage)cm.Item).certificate = new CertificateType(FLocalNickName,FLocalX509Cert);
 						CNSender.sendCM(cm);
 					}else if(FKnownUsers.Contains(gcm.receiver)
 					         && ((CUser)FKnownUsers[gcm.receiver]).PCert!=null){
 					         	chatmessage cm = new chatmessage(typeof(CertificateMessage),32,false);
 								((CertificateMessage)cm.Item).receiver = new String[] {gcm.sender} ;
 					         	((CertificateMessage)cm.Item).certificate = new CertificateType(gcm.receiver,((CUser)FKnownUsers[gcm.receiver]).PCert.RawData);
 					         	CNSender.sendCM(cm);
 					         }else sendFurther(pcm,gcm.receiver);
 					
 				}
 				
 			}
 			#endregion
 			#region Certificate
 			//just send to other receivers, if any (coz serializer already should have added cert)
 			else if (null != (certm = pcm.Item as CertificateMessage) ){
 				if (pcm.ttl == 0) Logger.Log("ConMan",2,"ttl reached 0",pcm);
 				else if (certm.receiver!=null){
 					ArrayList al = new ArrayList();
 					foreach (string mem in certm.receiver)
 						if (null!= ( (cu = FKnownUsers[mem] as CUser) ) )
 							al.Add(cu);
 					SendToUsers(al,pcm,0,false);
 				}
 			}
 			#endregion
 			#region getKey
 			else if (null != (gkm = pcm.Item as GetKeyMessage) ){
 				if ((gkm.receiver!=null) && gkm.receiver.Equals(FLocalNickName)){
 					if (null!= ( (cc = FKnownChannels[gkm.channel] as CChannel) ) )
 						CNSender.sendCM(cc.buildKM(gkm,(CUser)FKnownUsers[gkm.sender]));
 					else
 						//channel not known->create...
 						addToKnownChannels(gkm.channel);
 				}else sendFurther(pcm,gkm.receiver);
 				
 			}
 			#endregion
 			#region Key
 			else if (null != (km = pcm.Item as KeyMessage) ){
 				if (km.receiver.Equals(FLocalNickName)){
 					if (null!= ( (cc = FKnownChannels[km.channel] as CChannel) ) )
 						cc.PSymKey = CEncryption.AsymDecrypt(km.key,PLocalRSAM);
 					else
 						//channel not known->create...
 						addToKnownChannels(km.channel);
 				}else sendFurther(pcm,km.receiver);
 				
 			}
 			#endregion
 			#region Routing
 			else if (null != (rm = pcm.Item as RoutingMessage) ){
 				if (rm.destination!=null) foreach (DestinationType dt in rm.destination){
 					if (FKnownUsers.Contains(dt.user))
 						((CUser)FKnownUsers[dt.user]).addRoute(CNSender,(dt.hops+1));
 					else FKnownUsers.Add(dt.user,new CUser(dt.user,CNSender,(dt.hops+1)));
 					onNetInfoChanged(FKnownUsers[dt.user],NetInfoEventType.user);
 				}
 			}
 			#endregion
 			#region Message
 			else if (null != (mm = pcm.Item as MessageMessage) ){
 				if (mm.channel==null) mm.channel="";
 				if (null!= ( (cc = FKnownChannels[mm.channel] as CChannel) ) )
 					if (cc.PbLocalUserMember) cc.processMM(mm,CNSender);
 				else	//channel not known->create...
 				addToKnownChannels(mm.channel);
 				
 				//and resend it, if there are more people on the list...
 				if (pcm.ttl == 0) Logger.Log("ConMan",2,"ttl reached 0",pcm);
 				else if (mm.receiver!=null){
 					ArrayList al = new ArrayList();
 					foreach (string mem in mm.receiver){
 						if (!mem.Equals(PLocalNickName)){
 							if ((null!= ( (cu = FKnownUsers[mem] as CUser) ) )  )
 								al.Add(cu);
 							else addToKnownUsers(mem);
 						}
 					}
 					SendToUsers(al,pcm,0,false);
 				}else	//flood anon MMCM further...
 				floodCM(pcm,CNSender);
 				
 			}
 			#endregion
 			#region Obscure
 			else if (null != (om = pcm.Item as ObscureMessage) ){
 				if (om.receiver.Equals(FLocalNickName)){
 					//ok.. try to decrypt it..
 					byte[] buffer=null;
 					buffer = CEncryption.AsymDecrypt(om.text,PLocalRSAM);
 					chatmessage[] cma = CSerializer.parse(buffer);
 					
 					if (cma[0].Item is ObscureMessage){
 						//another one, send it to next station
 						ObscureMessage obscm = (ObscureMessage)cma[0].Item;
 						//local user another time?
 						if (obscm.receiver.Equals(PLocalNickName)){
 							processCM(cma[0],CNSender);
 						}else{
 							if (FKnownUsers.Contains(obscm.receiver))
 								((CUser)FKnownUsers[obscm.receiver]).PBestRoute.sendCM(cma[0],false,true);  //insert MID, but dont sign
 								else
 									addToKnownUsers(obscm.receiver);
 						}
 						
 					}else if (cma[0].Item is MessageMessage){
 						//last link in chain -> post it "jeeehaaa"
 						cma[0].flood = true;
 						string mmt = (string) ((MessageMessage)cma[0].Item).text.Value;
 						if (mmt!=null && mmt!=""){
 							try{
 								foreach (CNode cn in FKnownNodes.Values)
 									if (cn.PbLinked && cn!=PLocalNode)
 										cn.sendCM(cma[0],false,true);
 							}catch (Exception e){
 								Logger.Log("ConMan",1,"flooding anon MMCM",e);
 							}
 							//and post to local "" channel
 							((CChannel)FKnownChannels[((string)"")]).processMM((MessageMessage)cma[0].Item,CNSender);
 						}else Logger.Log("ConMan",5,"got Obscure Msg with void MessageMessage",cma[0]);
 					}
 					
 				}else sendFurther(pcm,om.receiver);
 				
 				
 			}
 			#endregion
 			else{
 				//hmm.. no known message?
 				Logger.Log("ConMan",1,"msg type not recognized",pcm);
 			}
 			
 			//see if we can learn something about routing from this msg
 			if ((pcm.Item is ASignedMessage) && !(pcm.Item is ObscureMessage)
 				&& ((ASignedMessage)pcm.Item).sender!=null
 				&& ((ASignedMessage)pcm.Item).sender!=FLocalNickName){
 					if (null != (cu = FKnownUsers[((ASignedMessage)pcm.Item).sender] as CUser) )
 						cu.addRoute(CNSender,(uint)Math.Abs((int)32-pcm.ttl));
 				}
 				
 		}
 		#endregion
 		
 		#region Interface methods
 		public static void 			globalConnect(){
 			if (FConnCount>4) return;
 			Logger.Log("ConMan",3,"Connecting all");
 			int LowerBound = 10, count=0;
 			try{
 				while (count < Math.Min(4,FKnownNodes.Count-1)){
 					foreach (CNode cn in FKnownNodes.Values){
 						if (cn.FMetric==LowerBound && !cn.PbLinked){
 							cn.openLink();
 							count++;
 						}
 					}
 					if (--LowerBound < 0) break;
 				}
 				
 			}catch(Exception e){ Logger.Log("ConMan",2,"Error Connecting all",e);}
 			
 		}
 		public static void 			globalDisConnect(){
 			Logger.Log("ConMan",3,"Disconnecting all");
 			try{
 				foreach (CNode cn in FKnownNodes.Values) cn.closeLink();
 			}catch(Exception e){ Logger.Log("ConMan",2,"Error DisConnecting all",e);}
 			FConnCount=0;
 		}
 		
 		public static CNode		 	addNode(IPEndPoint endPoint){
 			Logger.Log("ConMan",3,"adding node "+endPoint);
 			return addToKnownNodes(endPoint);
 		}
 		
 		public static void 			remNode(INode node){
 			Logger.Log("ConMan",3,"removing node "+node);
 			FKnownNodes.Remove(node.PNodeName.ToString());
 		}
 		
 		public static IChannel		createChannel(string name){
 			CChannel cc = addToKnownChannels(name,false,true);
 			//have to send channel CM
 			floodCM(cc.buildChannelCM(),null);
 			return cc;
 		}
 		public static IChannel		createChannel(string name,bool pbClosed){
 			CChannel cc = addToKnownChannels(name,pbClosed,true);
 			floodCM(cc.buildChannelCM(),null);
 			return cc;
 		}
 		public static void			sendGetCertificateMsg(IUser User){
 			if (FbOnlyNode){
 				Logger.Log("ConMan",3,"cannot send signed GetCertCM in only NodeMode for "+User.PNick);
 				throw new ENetException("cannot get Cert in NodeOnly Mode");
 			}else{
 				if(!User.PbGCMPending){
 					Logger.Log("ConMan",3,"trying to get cert for "+User.PNick);
 					chatmessage cm = new chatmessage(typeof(GetCertificateMessage),32,false);
 					((GetCertificateMessage)cm.Item).receiver = User.PNick;
 					User.PBestRoute.sendCM(cm);
 					User.PbGCMPending=true;
 				}else{
 					Logger.Log("ConMan",3,"GetCertCM already pending for "+User.PNick+",accepting temp.");
 					User.PbGCMPending=false;
 					throw new ENetException("Cert already pending...");
 				}
 			}
 		}
 		public static void			sendCertificateMsgTo(string strUser){
 			if (FKnownUsers.Contains(strUser))
 				sendCertificateMsgTo((CUser)FKnownUsers[strUser]);
 		}
 		public static void			sendCertificateMsgTo(IUser User){
 			if (FbOnlyNode) Logger.Log("ConMan",3,"cannot send CertCM without local user to "+User.PNick);
 			else{
 				Logger.Log("ConMan",3,"sending local cert to "+User.PNick);
 				chatmessage cm = new chatmessage(typeof(CertificateMessage),32,false);
 					((CertificateMessage)cm.Item).receiver = new String[] {User.PNick} ;
 				((CertificateMessage)cm.Item).certificate = new CertificateType(FLocalNickName,FLocalX509Cert);
 				User.PBestRoute.sendCM(cm);
 			}
 		}
 		public static CUser			addToKnownUsers(string name){
 			if (name==null) return null;
 			if (FKnownUsers.Contains(name)) return (CUser)FKnownUsers[name];
 			CUser cu = new CUser(name);
 			FKnownUsers.Add(name,cu);
 			onNetInfoChanged(cu, NetInfoEventType.user);
 			return cu;
 		}
 		public static CChannel		addToKnownChannels(string name){
 			return addToKnownChannels(name,false,false);
 		}
 		public static CChannel		addToKnownChannels(string name,bool pbClosed){
 			return addToKnownChannels(name,pbClosed,false);
 		}
 		public static CChannel		addToKnownChannels(string name,bool pbClosed,bool join){
 			if (name==null) return null;
 			if (FKnownChannels.Contains(name)) return (CChannel)FKnownChannels[name];
 			CChannel cc = new CChannel(name,pbClosed,join);
 			FKnownChannels.Add(name,cc);
 			onNetInfoChanged(cc, NetInfoEventType.channel);
 			return cc;
 		}
 		public static CChannel		addToKnownChannels(ChannelMessage chanm){
 			if (FKnownChannels.Contains(chanm.channel)) return (CChannel)FKnownChannels[chanm.channel];
 			CChannel cc = new CChannel(chanm);
 			FKnownChannels.Add(cc.PName,cc);
 			onNetInfoChanged(cc, NetInfoEventType.channel);
 			return cc;
 		}
 		public static CNode			addToKnownNodes(IPEndPoint ep){
 			if (ep==null) return null;
 			if (FKnownNodes.Contains(ep.ToString())) return (CNode)FKnownNodes[ep.ToString()];
 			CNode cn = new CNode(ep);
 			FKnownNodes.Add(ep.ToString(),cn);
 			cn.EvStateChanged+=new DNodeChangedEventHandler(NodeStateChanged);
 			onNetInfoChanged(cn, NetInfoEventType.node);
 			return cn;
 		}
 		public static CNode			changeLocalNode(IPEndPoint Oldep){
 			CNode cn = null;
 			try{
 				if (Oldep!=null){
 					if (FKnownNodes.Contains(FLocalEndPoint.ToString())) return (CNode)FKnownNodes[FLocalEndPoint.ToString()];
 					FKnownNodes.Remove(Oldep.ToString());
 				}
 				cn = new CNode(true);
 				FKnownNodes.Add(FLocalEndPoint.ToString(),cn);
 				//cn.EvStateChanged+=new DNodeChangedEventHandler(NodeStateChanged);
 				onNetInfoChanged(cn, NetInfoEventType.node);
 			}catch(Exception){ }
 			return cn;
 		}
 		public static CNode			addToKnownNodes(Socket s){
 			if (FKnownNodes.Contains(s.RemoteEndPoint.ToString())) return (CNode)FKnownNodes[s.RemoteEndPoint.ToString()];
 			CNode cn = new CNode(s);
 			FKnownNodes.Add(s.RemoteEndPoint.ToString(),cn);
 			cn.EvStateChanged+=new DNodeChangedEventHandler(NodeStateChanged);
 			onNetInfoChanged(cn, NetInfoEventType.node);
 			return cn;
 		}
 		#endregion
 		
 		#region public helper funcs
 		
 		
 		/* static Method floodCM
 		 * --------------------
 		 * (re)floods a message to all links
 		 *  drops it, if flood=false and ttl==0
 		 * doesnt send it to the node, where it came from
 		 * 
 		 * in:	CNode sender	- the sender (Node) from where the Msg came
 		 * in:	chatmessage pcm	- the ChatMessage to flood
 		*/
 		public static void floodCM(chatmessage pcm,CNode sender){
 			try{
 				if (pcm.flood && pcm.ttl > 0)
 					foreach (CNode cn in FKnownNodes.Values)
 						try{
 							
 							if (cn!=PLocalNode && cn.PbLinked && cn!=sender)
 								cn.sendCM(pcm,(sender==null));
 						}catch(Exception e){
 							Logger.Log("ConMan",1,"Error sending CM to Node",pcm,e);
 						}
 			}catch(Exception e){
 				Logger.Log("ConMan",1,"Error flooding CM",pcm,e);
 			}
 		}
 		
 		
 		
 		/* static Method SendToUsers
 		 * --------------------
 		 * Sends a CM to all Users in a Collection
 		 *  
 		 * in:	ICollection pToUsers 	- Collection of users (IUser) to send to
 		 * in:	chatmessage cm			- the ChatMessage to send
 		 * in:  int level				- the recursion level for error handling (aborts at > 3)
 		 * in:  bool pNew				- when the Msg is new, it needs to be signed
 		*/
 		public static void SendToUsers(ICollection pToUsers,chatmessage cm, int level,bool pNew){
 			if (pToUsers.Count == 0 && !(cm.Item is AMulticastMessage)) return;
 			if (++level>3) throw new ENetException("Routing table inconsistent and repair failed!");
 			ArrayList WaitingAl = new ArrayList(pToUsers);
 			if (WaitingAl.Contains(FLocalUser)) WaitingAl.Remove(FLocalUser);
 			ArrayList al = new ArrayList();
 			ArrayList RemAl = new ArrayList();
 			
 			foreach (CNode node in FKnownNodes.Values){
 				try{
 					
 					al.Clear();RemAl.Clear();
 					if(node.PbLinked){
 						foreach (CUser mem in WaitingAl){
 							try{
 								if (mem.PBestRoute == node){
 									al.Add(mem.PNick);
 									RemAl.Add(mem);
 								}
 							}catch(Exception e){
 								Logger.Log("ConMan",1,"Error sending to user"+mem.PNick+" omitting receiver",cm,e);
 							}
 						}
 						if (al.Count > 0){
 							((AMulticastMessage)cm.Item).receiver = (string[])al.ToArray(typeof(string));
 							node.sendCM(cm,pNew);
 						}
 						foreach (CUser mem in RemAl)
 							WaitingAl.Remove(mem);
 						
 					}
 				}catch (Exception e){
 					Logger.Log("ConMan",1,"Error while routing to node",cm,e);
 				}
 				
 			}
 			if (WaitingAl.Count > 0){
 				//try to rebuild routing table and retry...
 				Logger.Log("ConMan",2,"Inconsistent Routing table");
 				RepairRoutingTables();
 				SendToUsers(WaitingAl,cm,level++,pNew);
 			}
 		}
 		
 		/* static Method SendAnonMsg
 		 * --------------------
 		 * Sends an Anonymous Msg
 		 *  
 		 * in:  string pMsg	- the string to send
 		*/
 		public static void sendAnonMsg(string pMsg){
 			
 			ArrayList al = new ArrayList();
 			Random rd = new Random();
 			int count = rd.Next(3,6);
 			CUser[] cua = new CUser[FKnownUsers.Count];
 			Logger.Log("ConMan",3,"building anon msg with "+count+" links");
 			
 			FKnownUsers.Values.CopyTo(cua,0);
 			
 			while (al.Count<count){
 				int p = rd.Next(FKnownUsers.Count);
 				if (cua[p].PCert!=null)
 					al.Add(cua[p]);
 			}
 			//now we've got the list-> build the msgs
 			
 			
 			chatmessage cm = new chatmessage(typeof(MessageMessage),32,false);
 			chatmessage cm2=cm;
 			
 			MessageMessage mm = (MessageMessage)cm.Item;
 			mm.text = new MessageMessageText();
 			mm.text.Value=pMsg;
 			
 			//serialize the MM
 			byte[] interm;
 			try{
 				interm = CSerializer.build(cm,false);		//and dont sign
 				
 			}catch(Exception e){
 				Logger.Log("ConMan",4,"Error ser. MM CM",cm,e);
 				return;
 			}
 			CUser culast=null;
 			
 			foreach(CUser cu in al){
 				//do we have a cert 4 this user?
 				if (cu.PCert!=null){
 					Logger.Log("ConMan",4,"building obscure msg for "+cu.PNick);
 					
 					//encrypt
 					try{
 						interm = CEncryption.AsymEncrypt(interm, cu.PRSAM);
 					}catch(Exception e){
 						Logger.Log("ConMan",4,"Error while enc. obscure msg ",e);
 						throw new EEncryptingException("Error while enc. obscure msg:"+e);
 					}
 					
 					cm2 = new chatmessage(typeof(ObscureMessage),32,false);
 					((ObscureMessage)cm2.Item).receiver = cu.PNick;
 					((ObscureMessage)cm2.Item).text = interm;
 					
 					//serialize
 					try{
 						interm = CSerializer.build(cm2,false);
 					}catch(Exception e){
 						Logger.Log("ConMan",4,"Error while ser. obscure msg ",e);
 						throw new EXMLParsingException("Error while enc. obscure msg:"+e);
 					}
 					culast = cu;
 				}
 			}
 			//random ttl and send cm2 to last user in chain
 			cm2.ttl = rd.Next(28,33);
 			culast.PBestRoute.sendCM(cm2,false,true);		//dont sign but put in allhashed MID
 			//dont display to the local "" channel.. coz MM will be bounced back to us
 			//((CChannel)FKnownChannels[((string)"")]).processMM(mm,null);
 			
 		}
 		
 		#endregion
 		
 		#endregion
 		
 		#region Properties
 		#region Options
 		public static bool PbSendRandomAnonMsgs
 		{
 			get{
 				return FbSendRandomAnonMsgs;
 			}
 			set{
 				FbSendRandomAnonMsgs = value;
 			}
 		}
 		public static string PLocalNickName
 		{
 			get{
 				return FLocalNickName;
 			}
 			set{
 				if (value!=null && value!=""){
 					if (!FKnownUsers.Contains(value)){
 						FLocalNickName = value;
 						FLocalUser = new CUser(value,true);
 						FKnownUsers.Add(value,FLocalUser);
 						Logger.Log("ConMan",3,"Local NickName set to "+value);
 						if (FLocalRSA!=null && FLocalX509Cert!=null) FbOnlyNode = false;
 					}
 				}else{
 					FLocalNickName = null;
 					Logger.Log("ConMan",3,"Local NickName reset");
 					FbOnlyNode = true;
 				}
 			}
 		}
 		
 		public static int PLocalListenPort
 		{
 			get{
 				return FLocalListenPort;
 			}
 			set{
 				if(value>0 && value<65535){
 					FLocalListenPort = value;
 					IPEndPoint OldeP = FLocalEndPoint;
 					StartListening();
 					//and add local node to knownNodes
 					changeLocalNode(OldeP);
 				}
 			}
 		}
 		
 		public static RSA PLocalRSA
 		{
 			get{
 				return FLocalRSA;
 			}
 			set{
 				if (value!=null){
 					RSAParameters rp = value.ExportParameters(true);
 					if (rp.D!=null && rp.D.Length>0){
 						FLocalRSA = value;
 						Logger.Log("ConMan",3,"Local RSA set");
 						if (FLocalNickName!=null && FLocalX509Cert!=null) FbOnlyNode = false;
 					}
 				}else{
 					FLocalRSA = null;
 					Logger.Log("ConMan",3,"Local RSA reset");
 					FbOnlyNode = true;
 				}
 			}
 		}
 		public static RSAManaged PLocalRSAM{
 			get{
 				RSAManaged rsam = new RSAManaged();
 				rsam.ImportParameters(FLocalRSA.ExportParameters(true));
 				return rsam;
 			}
 		}
 		public static byte[] PLocalX509Cert
 		{
 			get{
 				return FLocalX509Cert;
 			}
 			set{
 				if (value!=null && value.Length > 0){
 					FLocalX509Cert = value;
 					Logger.Log("ConMan",3,"Local cert set");
 					((CUser)FKnownUsers[FLocalNickName]).PCert = new Mono.Security.X509.X509Certificate(value);
 					if (FLocalNickName!=null && FLocalRSA!=null) FbOnlyNode = false;
 				}else{
 					FLocalX509Cert = null;
 					Logger.Log("ConMan",3,"Local cert reset");
 					FbOnlyNode = true;
 				}
 			}
 		}
 		#endregion
 		
 		#region info
 		public static CUser PLocalUser
 		{
 			get{
 				return FLocalUser;
 			}
 		}
 		public static IPEndPoint PLocalNodeName
 		{
 			get{
 				return FLocalEndPoint;
 			}
 		}
 		
 		public static CNode PLocalNode
 		{
 			get{
 				return (CNode)FKnownNodes[FLocalEndPoint.ToString()];
 			}
 		}
 		
 		public static Hashtable PKnownChannels
 		{
 			get{
 				return FKnownChannels;
 			}
 		}
 		
 		public static Hashtable PKnownNodes
 		{
 			get{
 				return FKnownNodes;
 			}
 		}
 		
 		public static Hashtable PKnownUsers
 		{
 			get{
 				return FKnownUsers;
 			}
 		}
 		#endregion
 		#endregion
 		
 		
 		#region private funcs
 		
 		/* static Method SendFurther
 		 * -------------------------
 		 * Sends a Msg further to a User
 		 *  
 		 * in:  chatmessage pcm		- the CM to send
 		 * in:  string receiver		- the User to send the CM to
 		*/
 		private static void sendFurther(chatmessage pcm,string receiver){
 			try{
 				//have to send it further
 				if (pcm.ttl == 0) Logger.Log("ConMan",2,"ttl reached 0",pcm);
 				else{
 					if (FKnownUsers.Contains(receiver))
 						((CUser)FKnownUsers[receiver]).PBestRoute.sendCM(pcm,false);
 					else addToKnownUsers(receiver);
 				}
 			}catch (Exception e){
 				Logger.Log("ConMan",1,"error while sending CM further",pcm,e);
 			}
 		}
 		
 		/* static Method RepairRoutingTables
 		 * ----------------------------------
 		 * Repairs all routing tables for all users
 		 */
 		private static void RepairRoutingTables(){
 			try{
 				foreach (CUser cu in FKnownUsers.Values)
 					cu.RepairRouting();
 			}catch(Exception){}
 			Logger.Log("ConMan",3,"Repaired local routing tables");
 		}
 		
 		/* static Method buildRoutingCM
 		 * ------------------------------
 		 * Builds a routing Msg for a given Node
 		 *  
 		 * in:  CNode forNode - for this Node
 		*/
 		private static chatmessage buildRoutingCM(CNode forNode){
 			chatmessage cm = new chatmessage(typeof(RoutingMessage),1,false);
 			ArrayList al = new ArrayList();
 			
 			try{
 				foreach (CUser cu in FKnownUsers.Values)
 					if (cu.PBestRoute != forNode)
 						al.Add(new DestinationType(cu.PNick,cu.PBestRouteHops));
 			}catch(Exception){}
 			
 			((RoutingMessage)cm.Item).destination = (DestinationType[])al.ToArray(typeof(DestinationType));
 			Logger.Log("ConMan",4,"Built RoutingCM for "+forNode);
 			return (al.Count>0)?cm:null;
 		}
 		
 		/* static Method StartListening
 		 * --------------------
 		 * Starts the local listening Socket
 		*/
 		private static void StartListening(){
 			try{
 				if (FLocalListeningSocket!=null){
 					FLocalListeningSocket.Shutdown(SocketShutdown.Both);
 					FLocalListeningSocket.Close();
 				}
 			}catch(Exception){}
 			FLocalListeningSocket = null;
 			FLocalEndPoint = new IPEndPoint(IPAddress.Any,FLocalListenPort);
 			FLocalListeningSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
 			
 			try{
 				FLocalListeningSocket.Bind(FLocalEndPoint);
 				FLocalListeningSocket.Listen(10);
 			}catch(Exception e){
 				FLocalListeningSocket = null;
 				FLocalEndPoint = null;
 				Logger.Log("ConMan",1,"Listening on port "+FLocalListenPort+" failed:"+e);
 				throw new ENetException("Listening on port "+FLocalListenPort+" failed");
 			}
 			
 			//call BeginAccept to begin accepting asyncily
 			FLocalListeningSocket.BeginAccept(
 			                                  new AsyncCallback(NewInboundConnection),
 			                                  FLocalListeningSocket
 			                                  );
 			Logger.Log("ConMan",3,"Listening on port "+FLocalListenPort);
 			
 		}
 		
 		
 		/* static Method NewInboundConnection
 		 * --------------------
 		 * Gets called from the Socket when a New Connection is Inbound
 		 * creates a new CNode if necc. and sets its Link to the Socket
 		 * 
 		 * in:  IAsyncResult ar		- the asyncResult to see whether the call succeeded
 		*/
 		private static void NewInboundConnection(IAsyncResult ar){
 			Socket s;
 			//create new Node (if not alread there) with the newly made socket
 			try{
 				s = ((Socket)ar.AsyncState).EndAccept(ar);
 			}catch (Exception) { return; }
 			Logger.Log("ConMan",0,"Got connection from "+s.RemoteEndPoint);
 			
 			CNode cn;
 			if (null!= (cn = FKnownNodes[s.RemoteEndPoint.ToString()] as CNode) )
 				cn.openSocket(s);
 			else addToKnownNodes(s);
 			//FConnCount++;
 			//and re-begin accepting...
 			FLocalListeningSocket.BeginAccept(
 			                                  new AsyncCallback(NewInboundConnection),
 			                                  FLocalListeningSocket
 			                                  );
 		}
 		
 		
 		/* static Method NodeStateChanged
 		 * --------------------
 		 * Gets called when a NodeStateChanged Event is fired
 		 * 
 		 * in:  object sender			- the Node which called
 		 * in:  NetInfoEventType niet	- some Info about the Event
 		*
 		*/
 		private static void NodeStateChanged(object sender, EventArgs e){
 			onNetInfoChanged(sender, NetInfoEventType.node);
 			if (! ((CNode)sender).PbLinked) FConnCount--;
 			else FConnCount++;
 			
 			Logger.Log("ConMan",4,"state changed for node "+sender);
 			
 			if (!FbOnlyNode && FConnCount==0) globalConnect();
 		}
 		
 		/* static Method onNetInfoChanged
 		 * --------------------
 		 * Is called to fire the Event
 		 * Note: fires it asyncronously.. which is neccassary for the Sink might take a while to proccess
 		 * 
 		 * in:  object sender			- the Node which called
 		 * in:  NetInfoEventType niet	- some Info about the Event
 		*/
 		private static void onNetInfoChanged(object sender, NetInfoEventType niet){
 			Logger.Log("ConMan",3,niet+"-InfoChanged: "+sender);
 			NetInfoChangedEventArgs naet = new NetInfoChangedEventArgs(niet);
 			if (EvNetInfoChanged!=null)
 				//call the event async'ily (really cool from the .NET FW ;)
 				//coz events are really only derived from System.MulticastDelegate which is marked native (thus cant be overriden)
 				EvNetInfoChanged.BeginInvoke(sender,naet,null,null);
 		}
 		
 		/* static Method buildPeersCM
 		 * --------------------
 		 * Builds a PeersMessage
 		 * 
 		 * in:  CNode forNode		- for this Node
 		 * out: chatmessage			- the Message which has been built
 		*/
 		private static chatmessage buildPeersCM(CNode forNode){
 			chatmessage cm = new chatmessage(typeof(PeersMessage),1,false);
 			ArrayList al = new ArrayList();
 			
 			try{
 				foreach (CNode cn in FKnownNodes.Values)
 					if (cn!=PLocalNode && cn!=forNode)
 						al.Add(new PeerAddressType(cn.PNodeName.Address.ToString(),ItemChoiceType.ipaddress,cn.PNodeName.Port));
 			}catch(Exception){}
 			
 			((PeersMessage)cm.Item).peer = (PeerAddressType[])al.ToArray(typeof(PeerAddressType));
 			Logger.Log("ConMan",4,"Built PeersCM for "+forNode);
 			return (al.Count>0)?cm:null;
 		}
 		
 		
 		/* static Method OnTimerEvent
 		 * --------------------
 		 * Is the Timer Func that gets called every 10 Secs
 		 * Builds Channel Msgs for joined Channels, which hasnt seen CH Msgs for > 50 Secs
 		 * Builds routing Msgs for all linked Nodes
 		 * deletes old MIDs from the buffer
 		 * and also deletes stale Channels.. which havent seen any Msgs for > 3 Min
 		 * also sends eradically "" Anon Msgs if set
 		 * 
 		 * in:  object source		- the source (not used)
 		 * in:  ElapsedEventArgs e	- the Args to the the call (not used)
 		*/
 		private static void OnTimerEvent(object source, ElapsedEventArgs e)
 		{
 			#region CHANNEL
 			
 			if (!FbOnlyNode){
 				//send CHANNEL CMs
 				//build array with channel msgs to send
 				ArrayList al = new ArrayList();
 				
 				try{
 					foreach (CChannel cc in FKnownChannels.Values)
 						if (cc.PName!="" && cc.PbLocalUserMember &&
 						    (((TimeSpan) (DateTime.Now - cc.PLastChannelMessage) ).Seconds > 50) )
 						    al.Add(cc.buildChannelCM());
 				}catch (Exception ex){ Logger.Log("ConMan",5,"Exc while buidling CH CMs",ex); }
 				
 				//and send
 				try{
 					foreach (chatmessage cm in al)
 						floodCM(cm,null);
 				}catch(Exception ex){ Logger.Log("ConMan",5,"Exc while sending CH CMs",ex); }
 			}
 			#endregion
 			
 			#region ROUTING
 			
 			try{
 				foreach (CNode cn in FKnownNodes.Values)
 					if (cn!=PLocalNode && cn.PbLinked)
 						//send ROUTING CM for this link
 						//node takes care of not too flod its link with RM
 						cn.sendCM(buildRoutingCM(cn));
 				
 			}catch(Exception ex){Logger.Log("ConMan",5,"Exc while sending Routing CMs",ex);}
 			#endregion
 			
 			//delete some old entries from knownMsgs
 			CKnownMsgs.deleteOld();
 			
 			//remove stale channels...
 			try{
 				Hashtable ht = (Hashtable)FKnownChannels.Clone();
 				foreach (CChannel cc in ht.Values)
 					if ( !(cc.PName=="") && ((TimeSpan) (DateTime.Now - cc.PLastChannelMessage) ).Minutes > 3){
 						FKnownChannels.Remove(cc.PName);
 						Logger.Log("ConMan",5,"Deleted stale CH "+cc.PName);
 					}
 			}catch(Exception ex){Logger.Log("ConMan",5,"Exc while deleting stale CH",ex);}
 			
 			
 			//and send randomly some anon msgs
 			if (FbSendRandomAnonMsgs)
 				if (new Random().Next(100) > 94 )		//only send in 5% of all calls
 				sendAnonMsg("");
 			Logger.Log("ConMan",5,"Timer fired");
 			
 		}
 		
 		#endregion
 		
 		
 		#region private class for XML Peer Info de-/serializing
 		
 		[System.Xml.Serialization.XmlRootAttribute("peers-info", Namespace="http://www.ibr.cs.tu-bs.de/peers-info", IsNullable=false)]
 		public class PeersInfo {
 			
 			[System.Xml.Serialization.XmlElementAttribute("peer")]
 			public NodeSerBase[] peer;
 			
 			public PeersInfo(){ }
 			public PeersInfo(ICollection nodes){
 				//remove local node.. doesnt make sense...
 				ArrayList al = new ArrayList(nodes);
 				if (al.Contains(PLocalNode)) al.Remove(PLocalNode);
 				ArrayList al2 = new ArrayList();
 				foreach (CNode cn in al) al2.Add(new NodeSerBase(cn.PNodeName.ToString(),cn.FMetric));
 				peer = (NodeSerBase[])al2.ToArray(typeof(NodeSerBase));
 			}
 		}
 		
 		#endregion
 		
 		public class NodeSerBase {
 			public string 		Address;
 			public int			Metric;
				public NodeSerBase(){ }
				public NodeSerBase(string a, int m){
					Address=a;
					Metric=m;
				}
 		}
 		
	}
}
