/* CLink - class that represents a link
 * is only there if a Node is linked
 * takes care of sending and receiving, the proto stuff and also serializing them
 * has event for ConnLost
 *
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 03.07.2004
 * Time: 18:58
 *
 */
 using System;
 using System.Collections;
 using System.Net;
 using System.Threading;
 using System.Net.Sockets;
 using System.Security.Cryptography;
 using System.Text;

 using P2Pc.io;
 using P2Pc.util;

 namespace P2Pc.net{

 	public delegate void DLinkChangedEventHandler(object sender,
 	                                              EventArgs e);


 	public class CLink : ILink{
 		#region Fields

 		private Socket 		FSocket;
 		private volatile	int 		FiRx;
 		private volatile	int 		FiSx;
 		private DateTime 	FConnectedSince;
 		private IPEndPoint	FRemoteEndPoint;
 		private Thread		FThisThread;
 		//father node
 		private CNode		FFatherNode;
 		private RNGCryptoServiceProvider	FRNGcsp;
 		private ASCIIEncoding				FAasciiEnc;
 		private bool 		FbStayConnected=true;
 		#endregion

 		#region Events
 		public event DLinkChangedEventHandler EvStateChanged;
 		#endregion

 		#region Constructors
 		public CLink(){ }

 		public CLink(Socket ps, CNode pFather){
 			FFatherNode = pFather;
 			FRNGcsp = new RNGCryptoServiceProvider();
 			FAasciiEnc = new ASCIIEncoding();
 			//set the socket
 			FSocket = ps;
 			FConnectedSince = DateTime.Now;
 			FRemoteEndPoint = (IPEndPoint)ps.RemoteEndPoint;

 			//and start the thread ;)
 			FThisThread = new Thread(new ThreadStart(ThreadProc));
 			FThisThread.Start();

 			Logger.Log("Link:"+FRemoteEndPoint,4,"connected!");
 			//call the Event manually... coz it isnt reg. yet...
 			pFather.ConnStateChanged(this,null);

 		}

 		public CLink(IPEndPoint pEP, CNode pFather){
 			FAasciiEnc = new ASCIIEncoding();
 			FRNGcsp = new RNGCryptoServiceProvider();
 			FFatherNode = pFather;
 			FRemoteEndPoint = pEP;
 			FSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp );
 			FSocket.BeginConnect(pEP, new AsyncCallback(this.ConnectedCB), FSocket);

 		}

 		~CLink(){
 			try{
 				Close();
 			}catch(Exception){ }
 		}
 		#endregion


 		#region Methods

 		public void Close(){
 			try{
 				FbStayConnected = false;
 				FThisThread.Abort();
 			}catch(Exception e){
 					Logger.Log("Link:"+FRemoteEndPoint,5,"Error closing (already closed...)",e);
 			}
 			//dont join ;) brings us in a deadlock... coz were maybe already called out of this thread...
 			//and joining on ourself isnt a good idea ;)
 			//send quit msg only if socket still connected
 			if (PbConnected){
 				chatmessage cm = new chatmessage(typeof(QuitMessage),1);
 				((QuitMessage)cm.Item).reason = "Its a good day to die!";
 				try{
 					this.sendCM(cm,true,true);
 					FSocket.Shutdown(SocketShutdown.Both);
					FSocket.Close();
 				}catch(Exception e){
 					Logger.Log("Link:"+FRemoteEndPoint,5,"Error closing (already closed...)",e);
 				}

 			}
 			FSocket = null;
 			Logger.Log("Link:"+FRemoteEndPoint,0,"closed!");
 		}

 		public void sendCM(chatmessage pcm, bool pSign, bool pMID){

 			//put in MID
 			string mid="";
			if (pMID){
				 mid = CreateMID(pcm.Item is ObscureMessage);
 			}
 			if (pcm.Item is ANamedMessage) ((ANamedMessage)pcm.Item).messageid = mid;
			else if (pcm.Item is ASignedMessage) ((ASignedMessage)pcm.Item).messageid = mid;
			else if (pcm.Item is AMulticastMessage) ((AMulticastMessage)pcm.Item).messageid = mid;
			else if (pcm.Item is AUnicastMessage) ((AUnicastMessage)pcm.Item).messageid = mid;


 			byte[] toSend = null;
 			//serialize the msg
 			try{
 				toSend = CSerializer.build(pcm,pSign);
 			}catch(Exception de){
 				Logger.Log("Link:"+FRemoteEndPoint,1,"Error sirializing",pcm,de);
 				return;
 			}

 			try{
 				//send the length
 				FiSx += FSocket.Send(FAasciiEnc.GetBytes(toSend.Length.ToString()));
 				Logger.Log("Link:"+FRemoteEndPoint,4,"Sent size:"+toSend.Length.ToString());
 				//send CR/LF
 				FiSx += FSocket.Send(new byte[] {13,10} );

 				//and send the msg itself ;) blocks...
 				FiSx += FSocket.Send(toSend);
 				Logger.Log("Link:"+FRemoteEndPoint,4,"sent CM",pcm);
 				CKnownMsgs.AddMID(mid);

 				//add to the buffer of sent msgs
 			}catch (SocketException se){
 				if (!FSocket.Connected){
 					if (EvStateChanged!=null) EvStateChanged(this,null);
 					Logger.Log("Link:"+FRemoteEndPoint,2,"Send on Socket failed (maybe closed)",se);
 				}
 				Logger.Log("Link:"+FRemoteEndPoint,1,"Send on Socket failed",se);
 			}
 			//only makes sense to add flooded msgs


 		}

 		public override string ToString(){
 			return "Link{"+this.FRemoteEndPoint.ToString()+","+FiRx+","+FiSx+"}";
 		}

 		#region private funcs

 		/* Method ThreadProc
 		 * --------------------
 		 * Is the main Thread Method, which is execed in a Thread
 		 * Synchronously tries to receive something from the Socket
 		 * checks the Length and receives the Msg...
 		 * and send it to the Parent Node to process (after desirializing)
 		 *
 		 */
 		private void ThreadProc(){
 			byte[] sizeBuffer = new byte[16];
 			byte[] msgBuffer=null;
 			int i,toRecv,recvd;
 			byte lastByte;
 			chatmessage[] cma = null;

 			//is running in a thread so its save to sync'ily receive
 			while(PbConnected && FbStayConnected){
				try{

 					i=toRecv=recvd=0;
 					lastByte=0;cma=null;

 					//wait for CR
 					while(lastByte != 13 && lastByte<58){
 						FSocket.Receive(sizeBuffer,i,1,SocketFlags.None);
 						this.FiRx++;
 						lastByte = sizeBuffer[i];
 						i++;
 					}
 					if (lastByte==13){
 						//got CR, now get LF
 						FSocket.Receive(sizeBuffer,i,1,SocketFlags.None);
 						FiRx++;

 						if (sizeBuffer[i] == 10){
 							try{
 								toRecv = Int32.Parse(FAasciiEnc.GetString(sizeBuffer,0,i-1));
 							}catch(Exception aex){ Logger.Log("Link:"+FRemoteEndPoint,3,"Error ASCII-parsing Received size",aex);}
 							Logger.Log("Link:"+FRemoteEndPoint,4,"Received size:"+toRecv);

 							//now we have the int we have to recv in toRecv
 							if (toRecv > 0){
 								msgBuffer = new byte[toRecv];
 								//try to receive in loop
 								while (recvd < toRecv)
 									recvd += FSocket.Receive(msgBuffer,recvd,(toRecv-recvd),SocketFlags.None);

 								FiRx+=recvd;
 								if (recvd==toRecv){
 									//now all data should be there... try to parse it
 									try{
 										cma = CSerializer.parse(msgBuffer);
 									}catch(Exception e){
 										Logger.Log("Link:"+FRemoteEndPoint,1,"Error parsing:",e);
 									}
 									try{

 										//and now process it ;)
 										if (cma!=null) foreach(chatmessage cm in cma){
 											//check if its already known
											string mid="";
											if (cm.Item is ANamedMessage) mid = ((ANamedMessage)cm.Item).messageid;
				 							else if (cm.Item is ASignedMessage) mid = ((ASignedMessage)cm.Item).messageid;
											else if (cm.Item is AMulticastMessage) mid = ((AMulticastMessage)cm.Item).messageid;
											else if (cm.Item is AUnicastMessage) mid = ((AUnicastMessage)cm.Item).messageid;

 											if (!CKnownMsgs.isKnown(mid)){
 												CKnownMsgs.AddMID(mid);
 												Logger.Log("Link:"+FRemoteEndPoint,4,"got CM",cm);
 												//if not hand to father node
 												FFatherNode.processCM(cm,this);
 											}else
 												Logger.Log("Link:"+FRemoteEndPoint,4,"got known mID",cm);
 										}
									}catch(Exception e){
 										Logger.Log("Link:"+FRemoteEndPoint,1,"Error processing",e);
 									}

 								}
 							}
 						}
 					}


				}catch ( ThreadAbortException){
 					return;
 				}catch (Exception se){
 					if (!FSocket.Connected && FbStayConnected){
 						//conection lost, raise event
 						if (EvStateChanged!=null) EvStateChanged(this,null);
 						Logger.Log("Link:"+FRemoteEndPoint,1,"Recv on Socket failed (maybe closed)",se);
 						//and return... kills the thread
 						return;
					}
 				}

 			}//end while...
 		}

 		/* Method CreateMID
 		 * --------------------
 		 * Creates a MessageID
 		 *
 		 * in:  bool bAnon		- is a Anon Msg? Need to exclude local NickName and Date
 		 */
 		private string CreateMID(bool bAnon){

 			//create MID with secure random number gen...
 			byte[] srn = new byte[64];
 			string trail ="";
 			if (!bAnon) trail = "_"
 				+ ( (CConMan.PLocalNickName!=null)?(CConMan.PLocalNickName+"_"):"" )
 				+ DateTime.Now.ToString("dd.MM.yy_hh-mm-ss") ;


 			FRNGcsp.GetBytes(srn);

 			string ret="";
 			foreach (byte b in srn) ret+=b.ToString();

 			return ret+trail;
 		}


 		/* Method ConnectedCB
 		 * --------------------
 		 * Is the CallBack for the async Connect Call
 		 * Checks if succeeeded and then starts the thread
 		 *
 		 * in:  IAsyncResult ar		- the result...
 		 */
 		private void ConnectedCB(IAsyncResult ar){
 			if (ar.IsCompleted && PbConnected){
 				//Connected so set the Field
 				FConnectedSince = DateTime.Now;
 				//and start the thread ;)
 				FThisThread = new Thread(new ThreadStart(ThreadProc));
 				FThisThread.Start();
 				Logger.Log("Link:"+FRemoteEndPoint,0,"connected!");
 			}
 			if (EvStateChanged!=null) EvStateChanged(this,null);

 		}
 		#endregion
 		#endregion

 		#region Properties
 		public bool PbConnected
 		{
 			get{
 				return( (FSocket!=null)?FSocket.Connected : false );
 			}

 		}

 		public IPEndPoint PRemoteEndPoint{
 			get{
 				return FRemoteEndPoint;
 			}
 		}
 		public DateTime PConnectedSince
 		{
 			get{
 				return FConnectedSince;
 			}
 		}

 		public int PTrafficRx
 		{
 			get{
 				return FiRx;
 			}
 		}

 		public int PTrafficSx
 		{
 			get{
 				return FiSx;
 			}
 		}
 		/*
 		//in current bytes/sec
 		int PThroughput
 		{
 			get;
 		}
 		*/

 		public Socket PSocket
 		{
 			get{
 				return FSocket;
 			}
 		}
 		#endregion

 	}
}
