/* CNode - class that represents a Node
 * is there for every Known Node
 * just creates/destroys CLink objects and is a intermediary for incoming Msgs
 * checks HelloMsgs and reacts on QuitMsgs
 *
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 02.07.2004
 * Time: 17:20
 *
 */
 using System;
 using System.Collections;
 using System.Net;
 using System.Net.Sockets;
 using System.Xml.Serialization;

 using P2Pc.io;
 using P2Pc.util;

 namespace P2Pc.net{

 	public delegate void DNodeChangedEventHandler(object sender,EventArgs e);

 	public class CNode : INode{

 		#region Fields
 		[XmlIgnoreAttribute()]
 		private CLink		FLink;
 		[XmlIgnoreAttribute()]
 		private DateTime	FLastRoutingCM;
 		[XmlIgnoreAttribute()]
 		private DateTime	FLastRecvCM;
 		[XmlIgnoreAttribute()]
 		private IPEndPoint	FRemoteEndPoint;
 		private bool		FbCreating=false;

 		public bool 		FbLocal =	false;
 		public int			FMetric;
 		#endregion

 		#region Constructors



 		public CNode() {
 			FLastRecvCM = DateTime.Now;
 			FLastRoutingCM = DateTime.Now;
 		}

 		public CNode(Socket ps) : this() {
 			if (ps.Connected){
 				FbCreating=true;
 				FLink = new CLink(ps,this);
 				FLink.EvStateChanged += new DLinkChangedEventHandler(ConnStateChanged);
 				FRemoteEndPoint = (IPEndPoint)ps.RemoteEndPoint;
 				FMetric = 5;
 				FbCreating=false;
 				//if (EvStateChanged!=null) EvStateChanged(this,null);
 			}else throw new ENetException("Socket not connected");

 		}
 		public CNode(IPEndPoint pEP,int pM) : this(){
 			FRemoteEndPoint = pEP;
 			FMetric = pM;
 		}
 		public CNode(IPEndPoint pEP) : this(pEP,5) { }

 		public CNode(bool blocal) : this() {
 			if (blocal){
 				FbLocal=true;
 				FRemoteEndPoint = CConMan.PLocalNodeName;
 				FMetric = 10;
 			}
 		}

 		~CNode(){
 			//force disposing of the link
 			FLink = null;
 		}
 		#endregion

 		#region Events
 		public event DNodeChangedEventHandler EvStateChanged;
 		#endregion


 		#region Methods
 		public void sendCM(chatmessage pcm){
 			this.sendCM(pcm,true);
 		}
 		public void sendCM(chatmessage pcm, bool pNew){
 			sendCM(pcm,pNew,pNew);
 		}
 		public void	sendCM(chatmessage msg, bool pSign, bool pMID){
 			if (msg==null)
 				Logger.Log("Node:"+FRemoteEndPoint,2,"tried to send nullCM");
 			else if(FbLocal) CConMan.processCM(msg,this);
 			else if (PbLinked){
 				//see if its a ROUTING CM
 				if (msg.Item is RoutingMessage &&
 				    ( ((TimeSpan)(DateTime.Now - FLastRoutingCM) ).Seconds < 1) ){
 					//too short intervall
 						Logger.Log("Node:"+FRemoteEndPoint,2,"Too short intervall 4 RoutingCM",msg);
 						return;
 				}else FLastRoutingCM = DateTime.Now;

 				try{
 					FLink.sendCM(msg,pSign,pMID);
 				}catch(Exception ne){
 					Logger.Log("Node:"+FRemoteEndPoint,1,"Send to Node failed:"+ne,msg);
 					if (FLink!=null) FLink.Close();
					FLink = null;
 				}

 				if (EvStateChanged!=null) EvStateChanged(this,null);
 			}else Logger.Log("Node:"+FRemoteEndPoint,2,"tried to send CM to not linked node",msg);

 		}

 		public void openSocket(Socket ps){
 			FLink = new CLink(ps,this);
 			FLink.EvStateChanged += new DLinkChangedEventHandler(ConnStateChanged);
 			if (EvStateChanged!=null) EvStateChanged(this,null);
 		}

 		public void	openLink(){
 			if(FbLocal) return;
 			FLink = new CLink((IPEndPoint)FRemoteEndPoint,this);
 			FLink.EvStateChanged += new DLinkChangedEventHandler(ConnStateChanged);

 		}
 		public void	closeLink(){
 			if(FbLocal) return;
 			else if (FLink != null){
 				FLink.Close();
 				FLink = null;
 				if (EvStateChanged!=null) EvStateChanged(this,null);
 				Logger.Log("Node:"+FRemoteEndPoint,4,"closed link");
 			}
 		}

 		public void processCM(chatmessage pcm, IAbstractNetInfo sender){
 			//is called from below (i.e. the link)
 			FLastRecvCM = DateTime.Now;
 			HelloMessage hm;
 			//QuitMessage qm;
 			Logger.Log("Node:"+FRemoteEndPoint,5,"processing CM",pcm);

 			if (null != (hm = pcm.Item as HelloMessage) ){
 				//check version
 				if (hm.version != 2){
 					FLink = null;	//force close
 					if (EvStateChanged!=null) EvStateChanged(this,null);
 				}else{
 					CConMan.processCM(pcm,this);
 				}
 			}
 			else if (pcm.Item is QuitMessage ){
 				//FLink.Close();
 				FLink = null;	//force close
 				if (EvStateChanged!=null) EvStateChanged(this,null);
 			}else
 				CConMan.processCM(pcm,this);
 		}

 		public override string ToString(){
 			return "Node{" +( (FRemoteEndPoint!=null)?FRemoteEndPoint.ToString():"" )+",lnk="+PbLinked+",mtr="+FMetric+
				( (PbLinked&&FLink!=null) ? ",Rx="+FLink.PTrafficRx+",Sx="+FLink.PTrafficSx : "" )
				+ "}";
 		}



 		#region private methods

 		public void ConnStateChanged(object sender, EventArgs e){
 			if (FbCreating || (FLink!=null && FLink.PbConnected) ){
 				if (FbCreating) FLink=(CLink)sender;
 				//count up metric
 				if ((++FMetric) > 10) FMetric=10;
 				//send HelloMsg
 				chatmessage cm = new chatmessage(typeof(HelloMessage),1,false);
 				((HelloMessage)cm.Item).version = 2;
 				((HelloMessage)cm.Item).greeting = "Greetings from the third Planet!";
 				sendCM(cm,true);
 				//and send a GetPeers CM to get some new peers
 				sendCM(new chatmessage(typeof(GetPeersMessage),1,false),true);

 				Logger.Log("Node:"+FRemoteEndPoint,4,"Opened link");
 			}else{
 				if (FLink!=null) FLink.Close();
 				FLink = null;
 				//count down metric
 				if ((--FMetric) < -1) FMetric=-1;
 			}
 			if (EvStateChanged!=null) EvStateChanged(this,null);
 		}
 		#endregion
 		#endregion


 		#region Properties
 		public bool PbLinked
 		{
 			get{
 				return (FbLocal || ( (FLink != null)?FLink.PbConnected:false ) );
 			}
 			set{
 				if (value)	openLink();
 				else 		closeLink();
 			}
 		}

 		public CLink PLink
 		{
 			get{
 				return FLink;
 			}

 		}

 		public IPEndPoint PNodeName
 		{
 			get{
 				return (IPEndPoint)FRemoteEndPoint;
 			}
 		}

 		public DateTime PLastRecvCM
 		{
 			get{
 				return FLastRecvCM;
 			}
 		}

 		#endregion

 	}
}
