/* CUser - class that represents a User in the Network
 * is there for every known User
 * Just Holds the cert and the routing
 * 
 * Created by SharpDevelop.
 * User: Vitus
 * Date: 05.07.2004
 * Time: 13:53
 *
 */
 using System;
 using System.Collections;
 using Mono.Security.Cryptography;
 using Mono.Security.X509;

 using P2Pc.io;
 using P2Pc.util;

 namespace P2Pc.net{


 	public class CUser : IUser{
 		#region Fields
 		public bool FbGCMPending;

 		private RoutingTableEntry[]	FRouting;
 		private string				FNick;
 		private X509Certificate		FCert;

 		#endregion

 		#region Constructors

 		public CUser(string pNick){
 			FNick=pNick;
 			FRouting = new RoutingTableEntry[3];
 			for (int i=0;i<3;i++){ FRouting[i] = new RoutingTableEntry(); }

 		}

 		public CUser(string pNick, bool local) : this (pNick) {
 			if (local){
 				FRouting[0].FiHops = 0; //set to localhost
 				FRouting[0].FRouteOver = null;
 			}
 		}

 		public CUser(string pNick, CNode pRoute,uint pHops) : this (pNick) {
 			FRouting[0].FiHops = pHops;
 			FRouting[0].FRouteOver = pRoute;
 		}

 		#endregion

 		#region Methods

 		/*
 		public void sendPrivMsg(string theMsg){


 		}
 		*/
 		public void addRoute(CNode pRouteOver,uint pHops){
 			if (!pRouteOver.PbLinked) return;
			try{
 				if ( (pHops < FRouting[0].FiHops)
	 					|| ( !FRouting[0].FRouteOver.PbLinked )
					){
 					FRouting[0].FiHops = pHops;
 					FRouting[0].FRouteOver = pRouteOver;
 					Logger.Log("User:"+FNick,4,"added route[1] over "+pRouteOver+"("+pHops+")");
	 			}else if ( pRouteOver != FRouting[0].FRouteOver && 
						( (pHops < FRouting[1].FiHops)
 				        	  || ( !FRouting[1].FRouteOver.PbLinked ) )
					){
 					FRouting[1].FiHops = pHops;
 					FRouting[1].FRouteOver = pRouteOver;
	 				Logger.Log("User:"+FNick,4,"added route[2] over "+pRouteOver+"("+pHops+")");
 				}else if ( pRouteOver!=FRouting[0].FRouteOver && pRouteOver!=FRouting[1].FRouteOver &&
						( (pHops < FRouting[2].FiHops)
 						  || ( !FRouting[2].FRouteOver.PbLinked ) )
					){
 					FRouting[2].FiHops = pHops;
 					FRouting[2].FRouteOver = pRouteOver;
	 				Logger.Log("User:"+FNick,4,"added route[3] over "+pRouteOver+"("+pHops+")");
 				}
 			}catch(Exception){
 				Logger.Log("User:"+FNick,2,"Error adding new route");
 			}

 		}

 		public void RepairRouting(){
 			for (int i=2;i>=0;i--)			//if entry is there, but no longer linked, set the next best route
 				if ( (FRouting[i].FRouteOver!=null && !FRouting[i].FRouteOver.PbLinked) 
 						|| FRouting[i].FiHops>32 ){
 					
 					//copy next longer route
 					FRouting[i] = ( (i==2) ? (new RoutingTableEntry()) : FRouting[i+1] );
					//and delete it...
					if (i<2) FRouting[i+1]=new RoutingTableEntry();
				}
 			Logger.Log("User:"+FNick,4,"repaired routing");
 		}

 		public override string ToString(){
 			string Hop = (FRouting[0].FRouteOver!=null)? FRouting[0].FRouteOver.PNodeName.ToString() : "";
 			return "User{"+FNick+",crt="+(FCert!=null)+","+Hop+",hps="+FRouting[0].FiHops+"}";
 		}

 		#endregion

 		#region Properties

 		public CNode PBestRoute{
 			get{
 				if (FRouting[0].FiHops==0 && CConMan.PLocalNickName==FNick)
 					return CConMan.PLocalNode;
				else if (FRouting[0].FRouteOver!=null && FRouting[0].FRouteOver.PbLinked && FRouting[0].FiHops<33)
 					return FRouting[0].FRouteOver;
 				else{	//try to repair
					RepairRouting();
					if (FRouting[0].FRouteOver!=null && FRouting[0].FRouteOver.PbLinked && FRouting[0].FiHops<33)
 						return FRouting[0].FRouteOver;
					else throw new ENetException("User "+FNick+" not reachable!");
				}
 			}
 		}

 		public uint PBestRouteHops{
 			get{
 				return FRouting[0].FiHops;
 			}
 		}

 		public string PNick
 		{
 			get{
 				return FNick;
 			}
 		}

 		public bool PbGCMPending{
 			get{ return FbGCMPending; }
 			set{ FbGCMPending=value; }
 		}
 		public X509Certificate PCert
 		{
 			get{
 				return FCert;
 			}
 			set{
 				FCert = value;
 			}
 		}
 		public RSAManaged PRSAM{
 			get{
 				if (FCert==null) return null;
 				RSAManaged rsam = new RSAManaged();
 				rsam.ImportParameters(FCert.RSA.ExportParameters(false));
 				return rsam;
 			}
		}

 		#endregion

 		#region privates

 		private class RoutingTableEntry{
 			public CNode 	FRouteOver=null;
 			public uint	FiHops=Int32.MaxValue;

 		}

 		#endregion


 	}

}
